import typing as tp

import matplotlib.pyplot as plt
import numpy as np
import pandas as pd
from sklearn.preprocessing import OrdinalEncoder
from sklearn.tree import DecisionTreeRegressor, plot_tree
from sklearn.utils.validation import check_consistent_length
from sklift.metrics import uplift_by_percentile
from sklift.utils import check_is_binary

from auf.constants import RANDOM_STATE

plt.style.use("default")


def plot_uplift_by_percentile(
    y_true,
    uplift,
    treatment,
    strategy="overall",
    kind="line",
    bins=10,
    string_percentiles=True,
    axes=None,
    draw_bars="both",
):
    """Scikit-uplift implementation with 2 modifications:
        1. add axes parameter
        2. add flag for drawing only one of bar plots : with uplift or with response rates

    Args:
        y_true (1d array-like): Correct (true) binary target values
        uplift (1d array-like): Predicted uplift, as returned by a model
        treatment (1d array-like): Treatment labels
        strategy (string, ['overall', 'by_group']): Determines the calculating strategy. Default is 'overall'

            * ``'overall'``:
                The first step is taking the first k observations of all test data ordered by uplift prediction
                (overall both groups - control and treatment) and conversions in treatment and control groups
                calculated only on them. Then the difference between these conversions is calculated
            * ``'by_group'``:
                Separately calculates conversions in top k observations in each group (control and treatment)
                sorted by uplift predictions. Then the difference between these conversions is calculated

        kind (string, ['line', 'bar']): The type of plot to draw. Default is 'line'

            * ``'line'``:
                Generates a line plot
            * ``'bar'``:
                Generates a traditional bar-style plot

        bins (int): Determines а number of bins (and the relative percentile) in the test data. Default is 10
        string_percentiles (bool): type of xticks: float or string to plot. Default is True (string)
        axes: axes obtained from matplotlib.pyplot.subplots
        draw_bars (string, ['both', 'uplift', 'rates']): a parameter to be divided into bins and
            plotted on the graph. Default is 'both'

    Returns:
        Object that stores computed values.
    """

    strategy_methods = ["overall", "by_group"]
    kind_methods = ["line", "bar"]

    check_consistent_length(y_true, uplift, treatment)
    check_is_binary(treatment)
    check_is_binary(y_true)
    n_samples = len(y_true)

    if strategy not in strategy_methods:
        raise ValueError(
            f"Response rate supports only calculating methods in {strategy_methods},"
            f" got {strategy}."
        )

    if kind not in kind_methods:
        raise ValueError(
            f"Function supports only types of plots in {kind_methods}," f" got {kind}."
        )

    if not isinstance(bins, int) or bins <= 0:
        raise ValueError(f"Bins should be positive integer. Invalid value bins: {bins}")

    if bins >= n_samples:
        raise ValueError(
            f"Number of bins = {bins} should be smaller than the length of y_true {n_samples}"
        )

    if not isinstance(string_percentiles, bool):
        raise ValueError(
            f"string_percentiles flag should be bool: True or False."
            f" Invalid value string_percentiles: {string_percentiles}"
        )

    df = uplift_by_percentile(
        y_true,
        uplift,
        treatment,
        strategy=strategy,
        std=True,
        total=False,
        bins=bins,
        string_percentiles=False,
    )

    percentiles = df.index[:bins].values.astype(float)

    response_rate_trmnt = df.loc[percentiles, "response_rate_treatment"].values
    std_trmnt = df.loc[percentiles, "std_treatment"].values

    response_rate_ctrl = df.loc[percentiles, "response_rate_control"].values
    std_ctrl = df.loc[percentiles, "std_control"].values

    uplift_score = df.loc[percentiles, "uplift"].values
    std_uplift = df.loc[percentiles, "std_uplift"].values

    # uplift_weighted_avg = df.loc['total', 'uplift']

    check_consistent_length(
        percentiles,
        response_rate_trmnt,
        response_rate_ctrl,
        uplift_score,
        std_trmnt,
        std_ctrl,
        std_uplift,
    )

    if kind == "line":
        if axes is None:
            _, axes = plt.subplots(ncols=1, nrows=1, figsize=(8, 6))
        axes.errorbar(
            percentiles,
            response_rate_trmnt,
            yerr=std_trmnt,
            linewidth=2,
            color="forestgreen",
            label="treatment\nresponse rate",
        )
        axes.errorbar(
            percentiles,
            response_rate_ctrl,
            yerr=std_ctrl,
            linewidth=2,
            color="orange",
            label="control\nresponse rate",
        )
        axes.errorbar(
            percentiles,
            uplift_score,
            yerr=std_uplift,
            linewidth=2,
            color="red",
            label="uplift",
        )
        axes.fill_between(
            percentiles, response_rate_trmnt, response_rate_ctrl, alpha=0.1, color="red"
        )

        if np.amin(uplift_score) < 0:
            axes.axhline(y=0, color="black", linewidth=1)

        if string_percentiles:  # string percentiles for plotting
            percentiles_str = [f"0-{percentiles[0]:.0f}"] + [
                f"{percentiles[i]:.0f}-{percentiles[i + 1]:.0f}"
                for i in range(len(percentiles) - 1)
            ]
            axes.set_xticks(percentiles)
            axes.set_xticklabels(percentiles_str, rotation=45)
        else:
            axes.set_xticks(percentiles)

        axes.legend(loc="upper right")
        axes.set_title(f"Uplift by percentile\n")
        axes.set_xlabel("Percentile")
        axes.set_ylabel("Uplift = treatment response rate - control response rate")

    elif draw_bars == "both":  # kind == 'bar'
        delta = percentiles[0]

        # don't check axes if None : draw both plots on big figure
        fig, axes = plt.subplots(
            ncols=1, nrows=2, figsize=(8, 6), sharex=True, sharey=True
        )
        fig.text(
            0.04,
            0.5,
            "Uplift = treatment response rate - control response rate",
            va="center",
            ha="center",
            rotation="vertical",
        )

        axes[1].bar(
            np.array(percentiles) - delta / 6,
            response_rate_trmnt,
            delta / 3,
            yerr=std_trmnt,
            color="forestgreen",
            label="treatment\nresponse rate",
        )
        axes[1].bar(
            np.array(percentiles) + delta / 6,
            response_rate_ctrl,
            delta / 3,
            yerr=std_ctrl,
            color="orange",
            label="control\nresponse rate",
        )
        axes[0].bar(
            np.array(percentiles),
            uplift_score,
            delta / 1.5,
            yerr=std_uplift,
            color="red",
            label="uplift",
        )

        axes[0].legend(loc="upper right")
        axes[0].tick_params(axis="x", bottom=False)
        axes[0].axhline(y=0, color="black", linewidth=1)
        axes[0].set_title(f"Uplift by percentile\n")

        if string_percentiles:  # string percentiles for plotting
            percentiles_str = [f"0-{percentiles[0]:.0f}"] + [
                f"{percentiles[i]:.0f}-{percentiles[i + 1]:.0f}"
                for i in range(len(percentiles) - 1)
            ]
            axes[1].set_xticks(percentiles)
            axes[1].set_xticklabels(percentiles_str, rotation=45)

        else:
            axes[1].set_xticks(percentiles)

        axes[1].legend(loc="upper right")
        axes[1].axhline(y=0, color="black", linewidth=1)
        axes[1].set_xlabel("Percentile")
        axes[1].set_title("Response rate by percentile")

    elif draw_bars == "uplift":  # kind == 'bar'
        delta = percentiles[0]

        if axes is None:
            fig, axes = plt.subplots(
                ncols=1, nrows=1, figsize=(5, 5), sharex=True, sharey=True
            )
        fig.text(
            0.04,
            0.5,
            "Uplift = treatment response rate - control response rate",
            va="center",
            ha="center",
            rotation="vertical",
        )

        axes.bar(
            np.array(percentiles),
            uplift_score,
            delta / 1.5,
            yerr=std_uplift,
            color="red",
            label="uplift",
        )

        axes.legend(loc="upper right")
        axes.tick_params(axis="x", bottom=False)
        axes.axhline(y=0, color="black", linewidth=1)
        axes.set_title(f"Uplift by percentile\n")

    elif draw_bars == "rates":  # kind == 'bar'
        delta = percentiles[0]

        if axes is None:
            fig, axes = plt.subplots(
                ncols=1, nrows=1, figsize=(5, 5), sharex=True, sharey=True
            )
            fig.text(
                0.04,
                0.5,
                "Uplift = treatment response rate - control response rate",
                va="center",
                ha="center",
                rotation="vertical",
            )

        axes.bar(
            np.array(percentiles) - delta / 6,
            response_rate_trmnt,
            delta / 3,
            yerr=std_trmnt,
            color="forestgreen",
            label="treatment\nresponse rate",
        )
        axes.bar(
            np.array(percentiles) + delta / 6,
            response_rate_ctrl,
            delta / 3,
            yerr=std_ctrl,
            color="orange",
            label="control\nresponse rate",
        )

        if string_percentiles:  # string percentiles for plotting
            percentiles_str = [f"0-{percentiles[0]:.0f}"] + [
                f"{percentiles[i]:.0f}-{percentiles[i + 1]:.0f}"
                for i in range(len(percentiles) - 1)
            ]
            axes.set_xticks(percentiles)
            axes.set_xticklabels(percentiles_str, rotation=45)

        else:
            axes.set_xticks(percentiles)

        axes.legend(loc="upper right")
        axes.axhline(y=0, color="black", linewidth=1)
        axes.set_xlabel("Percentile")
        axes.set_title("Response rate by percentile")

    return axes


def plot_portrait_tree(
    x: pd.DataFrame,
    uplift: np.array,
    feature_names_dict: tp.Optional[tp.Dict[str, str]] = None,
    max_depth: int = 3,
    max_leaf_nodes: int = 6,
    min_samples_leaf: tp.Union[float, int] = 0.05,
    axes=None,
):
    """Builds a portrait of a high- and low-uplift clients using a decision tree

    Args:
        x (pd.DataFrame): sample with features on which the uplift model was trained
        uplift (1d array-like): predicted uplift, as returned by a model
        feature_names_dict (dict): dictionary of matching feature names, where
            keys are names of features in the data and values are names of
            features to be displayed in the plot
        max_depth (int): the maximum depth of the tree. Default is 3
        max_leaf_nodes (int): grow a tree with ``max_leaf_nodes`` in best-first
            fashion. Best nodes are defined as relative reduction in impurity. Default is 6
        min_samples_leaf (float or int): the minimum number of samples required
            to be at a leaf node. A split point at any depth will only be
            considered if it leaves at least ``min_samples_leaf`` training
            samples in each of the left and right branches. Default is 0.05

            - If int, then consider `min_samples_leaf` as the minimum number
            - If float, then `min_samples_leaf` is a fraction and
              `ceil(min_samples_leaf * n_samples)` are the minimum
              number of samples for each node
        axes: axes obtained from matplotlib.pyplot.subplots

    Idea:
        Train the decision tree on the same features as the uplift model,
        passing the uplift predicted by the uplift model as a target.
        Visualize the decision tree and see how it separates high- from
        low-uplift clients

    Returns:
        None
    """
    check_consistent_length(x, uplift)

    # encoder = OrdinalEncoder()
    # x_transformed = pd.DataFrame(encoder.fit_transform(x), columns=x.columns)

    # max_num_len = len(str(int(x_transformed.max(axis=None).max())))
    # x_transformed.fillna(int('9' * max_num_len), inplace=True)

    if isinstance(min_samples_leaf, float):
        min_samples_leaf = int(min_samples_leaf * len(x))

    if feature_names_dict:
        feats = [
            feature_names_dict[col] if col in feature_names_dict else col
            for col in x.columns
        ]
    else:
        feats = x.columns

    tree = DecisionTreeRegressor(
        max_depth=max_depth,
        max_leaf_nodes=max_leaf_nodes,
        min_samples_leaf=min_samples_leaf,
        random_state=RANDOM_STATE,
    )

    # tree.fit(x_transformed, uplift)
    tree.fit(x, uplift)

    fig = None
    if axes is None:
        fig, axes = plt.subplots(1, 1, figsize=(12, 8))

    plot_tree(tree, filled=True, feature_names=feats, ax=axes)

    plt.title("Client portrait: uplift depending on model features")
    plt.tight_layout()
    plt.show()
    return fig


def plot_uplift_by_feature_bins(
    feature: tp.Sequence[float],
    treatment: tp.Sequence[float],
    target: tp.Sequence[float],
    feature_name: str,
    amount_of_bins: int = 7,
    round_const: int = 2,
    axes=None,
):
    """Make plots of the distribution of the number of observations and mean target
       depending on the treatment type and feature bin

    Args:
        feature (1d array-like): feature values which values from 0.02 to 0.98
            percentile will be divided into bins and depending on the
            bin of which the graph is plotted
        treatment (1d array-like): treatment labels
        target (1d array-like): correct (true) binary target values
        feature_name (str, optional): feature name depending on the bin of
            which the graph is plotted
        amount_of_bins (int): number of bins into which the attribute will be
            divided. Default is 7
        round_const (int): number of digits after decimal point to which the
            values of numerical signs are rounded off. Default is 2
        axes: axes obtained from matplotlib.pyplot.subplots

    Returns:
        None
    """
    check_consistent_length(feature, treatment, target)

    if feature_name is None:
        feature_name = "feature"
    df = pd.DataFrame({feature_name: feature, "target": target, "treatment": treatment})

    amount_of_nans = df.loc[df[feature_name].isna()].shape[0]

    if df[feature_name].dtype == "object":
        category_counts = (
            df[feature_name]
            .value_counts(dropna=False)
            .nlargest(amount_of_bins)
            .index.tolist()
        )

        if df[feature_name].nunique() > amount_of_bins:
            category_counts.pop()

            if amount_of_nans > 0 and np.nan not in category_counts:
                category_counts.pop()

            category_counts.append(np.nan)
            category_counts.append("Other")

        df[feature_name] = df[feature_name].apply(
            lambda x: x if x in category_counts or pd.isna(x) else "Other"
        )

        if amount_of_nans > 0:
            df[feature_name] = (
                pd.Categorical(df[feature_name], categories=category_counts)
                .add_categories("NAN")
                .fillna("NAN")
            )
            df[feature_name] = df[feature_name].cat.reorder_categories(
                ["NAN"] + category_counts
            )
        else:
            df[feature_name] = pd.Categorical(
                df[feature_name], categories=category_counts
            )

    else:
        if amount_of_bins > df[feature_name].dropna().unique().shape[0]:
            df[feature_name] = df[feature_name].astype("object")
            sort_hist_by_object_cnt = False

        bins = np.linspace(
            df[feature_name].quantile(q=0.02),
            df[feature_name].quantile(q=0.98),
            amount_of_bins,
        )
        df = df.loc[
            (df[feature_name] >= df[feature_name].quantile(q=0.02))
            & (df[feature_name] <= df[feature_name].quantile(q=0.98))
            | (df[feature_name].isna())
        ]

        feature_cat = pd.cut(
            x=df[feature_name],
            bins=bins,
            precision=round_const,
            include_lowest=True,
            ordered=True,
        )

        interval_labels = [
            f"{intv.left:.{round_const}f} - {intv.right:.{round_const}f}"
            for intv in feature_cat.cat.categories
        ]

        category_mapping = {
            old: new for old, new in zip(feature_cat.cat.categories, interval_labels)
        }

        # Модифицированная часть: добавляем NAN только при необходимости
        if amount_of_nans > 0:
            df[feature_name] = (
                feature_cat.map(category_mapping)
                .cat.rename_categories(interval_labels)
                .cat.add_categories("NAN")
                .fillna("NAN")
            )
            categories_order = ["NAN"] + interval_labels
        else:
            df[feature_name] = feature_cat.map(category_mapping).cat.rename_categories(
                interval_labels
            )
            categories_order = interval_labels

        df[feature_name] = df[feature_name].cat.reorder_categories(categories_order)

    grouped = (
        df.groupby([feature_name, "treatment"])
        .agg(count=("target", "size"), mean_target=("target", "mean"))
        .reset_index()
    )

    if df[feature_name].dtype == "object":
        grouped.sort_values(by="count", inplace=True, ascending=False)

        if amount_of_nans > 0:
            nan_mask = grouped[feature_name] == "NAN"
            grouped = pd.concat(
                [grouped[nan_mask], grouped[~nan_mask]], ignore_index=True
            )

        other_mask = grouped[feature_name] == "Other"
        grouped = pd.concat(
            [grouped[~other_mask], grouped[other_mask]], ignore_index=True
        )

    else:
        grouped[feature_name] = pd.Categorical(
            grouped[feature_name],
            categories=df[feature_name].cat.categories,
            ordered=True,
        )
        grouped = grouped.sort_values(feature_name)

    treatment_1 = grouped[grouped["treatment"] == 1]
    treatment_0 = grouped[grouped["treatment"] == 0]

    fig = None
    if axes is None:
        fig, axes = plt.subplots(1, 2, figsize=(12, 5))

    ax1 = axes[0]
    width = 0.35  # Ширина столбцов
    x1 = np.arange(len(treatment_1[feature_name]))
    x0 = x1 + width

    ax1.bar(
        x1,
        treatment_1["count"],
        width,
        color="forestgreen",
        edgecolor="black",
        label="Treatment",
    )
    ax1.bar(
        x0,
        treatment_0["count"],
        width,
        color="orange",
        edgecolor="black",
        label="Control",
    )

    ax1.set_title("Treatment and control group sizes")
    # ax1s.set_xlabel("Bin")
    # ax1.set_xlabel(feature_name)
    ax1.set_ylabel("Number of observations")
    ax1.set_xticks(x1 + width / 2)
    ax1.set_xticklabels(treatment_1[feature_name])
    ax1.legend(loc="upper right")

    ax2 = axes[1]
    ax2.plot(
        treatment_1[feature_name],
        treatment_1["mean_target"],
        color="forestgreen",
        label="Treatment\ntarget rate",  # "Treatment 1",
        marker="o",
    )
    ax2.plot(
        treatment_0[feature_name],
        treatment_0["mean_target"],
        color="orange",
        label="Control\ntarget rate",  # "Treatment 0",
        marker="o",
    )
    ax2.set_title(f"Uplift")
    # ax2.set_xlabel("Bin")
    # ax2.set_xlabel(feature_name)
    # ax2.set_ylabel("Average target")
    ax2.legend(loc="upper right")

    fig.suptitle(f"Uplift and observations count by {feature_name} buckets")

    for ax in axes:
        ax.tick_params(axis="x", rotation=45)

    return fig



import typing as tp

import numpy as np
import pandas as pd
from sklift.metrics import uplift_by_percentile


class UpliftCalibration:
    """описание идеи работы класс
    1) храним модель и коэфффициенты по бакетам
    2) храним границы бакетов (при деградации переобучим все, с калибровками)
    3) для новых данных предсказания умножаем на коэфф бакета скора клиента
    4) для выборки можно требовать наличие всех фич модели и ни одной лишней
    """

    def __init__(
        self,
        model: tp.Any,
        features: tp.List[str],
        target_col: str,
        treatment_col: str,
        treatment_groups_mapper: tp.Dict[tp.Any, int] = {0: 0, 1: 1},
        bins: int = 10,
    ):
        self._model: tp.Any = model
        self._features: tp.List[str] = features
        self._target_col: str = target_col
        self._treatment_col: str = treatment_col
        self._treatment_groups_mapper: tp.Dict[tp.Any, int] = treatment_groups_mapper
        # self._treatment_groups_mapper_inv: tp.Dict[tp.Any, int] = {
        #     value: key for key, value in treatment_groups_mapper
        # }
        self._bins: int = bins
        self._bucket_coeffs: np.array = None
        self._uplift_borders_left: np.array = None  # desc
        self._uplift_borders_right: np.array = None  # desc

    def fit(self, data: pd.DataFrame):
        """Учим калибровочные коэффициенты на переданной выборке.

        Args:
            data (pd.DataFrame): выборка для подбора калибровочных коэффициентов
        """
        uplift = self._model.predict(data[self._features])
        target = data[self._target_col]
        treatment = data[self._treatment_col].map(self._treatment_groups_mapper)
        uplift_bucket_stats = [
            (np.min(bucket), np.max(bucket), np.mean(bucket))
            for bucket in np.array_split(np.sort(uplift)[::-1], self._bins)
        ]
        self._uplift_borders_left, self._uplift_borders_right, mean_uplift_preds = map(
            list, zip(*uplift_bucket_stats)
        )
        self._uplift_borders_left[-1] = -np.inf
        self._uplift_borders_right[0] = np.inf
        mean_uplift = uplift_by_percentile(target, uplift, treatment, bins=self._bins)[
            "uplift"
        ].values
        self._bucket_coeffs = mean_uplift / mean_uplift_preds

    def _get_coeffs(self, uplift: np.array):
        func1d = lambda num: [
            idx
            for idx in range(self._bins)
            if self._uplift_borders_left[idx] <= num
            and num <= self._uplift_borders_right[idx]
        ][0]
        func1d = np.vectorize(func1d)
        idxs = np.apply_along_axis(func1d, axis=-1, arr=uplift)
        return self._bucket_coeffs[idxs]

    def predict(self, data: pd.DataFrame):
        """Получение скоров для выборки с учетом калибровки.

        Args:
            data (pd.DataFrame): выборка для получения калиброванных скоров
        """
        uplift = self._model.predict(data[self._features])
        calibrated_uplift = self._get_coeffs(uplift) * uplift
        return calibrated_uplift



from .auf_tree import AufTreeClassifier
from .auf_forest import AufRandomForestClassifier
from .auf_x_learner import AufXLearner
from .auf_s_learner import AufSoloModel

__all__ = [
    'AufTreeClassifier',
    'AufRandomForestClassifier',
    'AufXLearner',
    'AufSoloModel'
]



import typing as tp
from dataclasses import dataclass

import numpy as np
import pandas as pd
from catboost import CatBoostClassifier, CatBoostRegressor
from causalml.inference.tree import UpliftRandomForestClassifier
from sklearn.pipeline import Pipeline
from sklift.models import SoloModel, TwoModels

from ..constants import (
    RANDOM_STATE,
    SOLO_MODEL,
    TWO_MODEL,
    UPLIFT_FOREST,
    UPLIFT_TREE,
    X_LEARNER,
)
from ..models import AufRandomForestClassifier, AufTreeClassifier, AufXLearner
from .gridsearch import ModelResult, OptuneUpliftDefault

# from functools import partical


def fit_model(
    estimator: Pipeline,
    df_train: pd.DataFrame,
    features: tp.List[str],
    target_col: str,
    treatment_col: str,
    # uplift_type: tp.Literal["abs", "rel"] = "abs"
    uplift_type: str = "abs",
):
    """
    Train the provided model on the training data and evaluate its performance on the validation data

    Args:
        estimator: model to be trained. It should be compatible with the scikit-learn interface

        df_train (pd.DataFrame): data for training the model. It should contain both features
            and the target variable

        df_valid (pd.DataFrame): data for validating the model. It is used to assess performance
            after training

        df_test (pd.DataFrame): data for testing the model. It is used for final evaluation of
            the model's performance

        features (1d array-like): list of feature names to be used for training the model

        target_col (string): name of the column representing the target variable that the model
            should predict

        treatment_col (string): name of the column representing treatment or group that should
            be considered during training

        uplift_type: type of uplift, if abs use TR_1-TR2 , if rel use TR_1/TR_2 -1

    Returns: dataclass ModelResults containing the trained model, its predictions, and the
             features on which it was trained
    """
    if type(estimator).__name__ == "CatBoostClassifier":
        estimator.fit(X=df_train[features], y=df_train[target_col])
    # elif type(estimator).__name__ == "UpliftRandomForestClassifier":
    #     estimator.fit(
    #         X=df_train[features].values,
    #         y=df_train[target_col].values,
    #         treatment=df_train[treatment_col].astype("str").values,
    #     )
    else:
        estimator.fit(
            X=df_train[features],
            y=df_train[target_col],
            treatment=df_train[treatment_col],
        )

    result = ModelResult(
        estimator=estimator,
        features=features,
        uplift_prediction_type=uplift_type,
        median_test_metric_value=None,
    )
    return result


def generate_model_from_classes(
    model_class: object,
    df_train: pd.DataFrame,
    df_valid: pd.DataFrame,
    df_test: pd.DataFrame,
    features: tp.Dict[str, tp.List[str]],
    target_col: str,
    treatment_col: str,
    feature_nums: tp.List[int],
    timeout: int,
    metric: object,
    use_default_params: bool,
    search_class: object,
    overfit_metric: object = None,
    # uplift_type: tp.Literal["abs", "rel"] = "abs"
    uplift_type: str = "abs",
):
    """
    Generate and train a model from the specified class, using the provided training and validation datasets

    Args:
        model_class: model class to be instantiated and trained

        df_train (pd.DataFrame): the dataset used for training the model. It should contain both features
            and the target variable

        df_valid (pd.DataFrame): the dataset used for hyperparameter tuning. It is used to evaluate the
            model's performance during training

        df_test (pd.DataFrame): the dataset used for final evaluation of the model's performance after training

        features (dict): a dictionary where each key corresponds to a specific feature set to be used
            for training the model

        target_col (string): the name of the column representing the target variable that the model should predict

        treatment_col (string): the name of the column representing treatment or group that should be considered
            during training

        feature_nums (1d array-like): an optional parameter specifying the number of features to use for
            training. If provided, the model will be trained using only this number of features

        timeout (integer): an optional parameter specifying the maximum time (in seconds) allowed for
            training if use_default_params is set to False

        metric: an optional parameter specifying the metric to optimize if use_default_params is set to False

        use_default_params (bool): a flag indicating whether to use default parameters for the model. If
            set to False, hyperparameter tuning will be performed

        search_class: the class responsible for hyperparameter tuning. It should implement methods to search
            for optimal parameters based on the provided metric

        overfit_metric: an optional parameter specifying the overfit_metric to optimize if use_default_params is set to False

        uplift_type: type of uplift, if abs use TR_1-TR2 , if rel use TR_1/TR_2 -1

    Returns: dict
        A dictionary containing lists of ModelResults for each model trained, indexed by their corresponding feature sets
    """
    fast_fit_result = dict()

    for key in features.keys():
        if feature_nums is None:
            feature_nums = [len(features[key])]

        result_n = list()

        for n in feature_nums:
            assert n <= len(
                features[key]
            ), f"Число признаков {n} в feature_nums, не может быть больше числа признаков {len(features[key])}"

            for uplift_type in ["propensity", "abs", "rel"]:

                if uplift_type == "rel" and model_class.__name__ in [
                    "CatBoostClassifier",
                    "UpliftRandomForestClassifier",
                    "AufXLearner",
                ]:
                    break

                if (
                    uplift_type == "propensity"
                    and model_class.__name__ != "CatBoostClassifier"
                ):
                    continue

                if use_default_params:
                    mc = get_default_model(model_class)
                else:
                    finder_class = search_class(
                        df_train,
                        df_valid,
                        metric,
                        treatment_col,
                        target_col,
                        overfit_metric,
                        uplift_type,
                    )

                    mc = finder_class.find_best_params(
                        model_class, features[key][:n], timeout
                    )

                result_n.append(
                    fit_model(
                        mc,
                        df_train,
                        features[key][:n],
                        target_col,
                        treatment_col,
                        uplift_type,
                    )
                )

        fast_fit_result[key] = result_n

    return fast_fit_result


def models_fast_fit(
    estimator: Pipeline,
    df_train: pd.DataFrame,
    df_valid: pd.DataFrame,
    df_test: pd.DataFrame,
    features: tp.Dict[str, tp.List[str]],
    target_col: str,
    treatment_col: str,
    feature_nums: tp.List[int],
):
    """
    Train specified model using several numbers of TOP features from each feature ranker.

    Args:
        estimator: the model to be trained

        df_train (pd.DataFrame): the dataset used for training the model. It should contain both features and the
            target variable

        df_valid (pd.DataFrame): the dataset used for hyperparameter tuning. It is used to evaluate the model's
            performance during training

        df_test (pd.DataFrame): the dataset used for final evaluation of the model's performance after training

        features (dict): a dictionary where each key corresponds to a specific feature set to be used for training
            the model

        target_col (string): the name of the column representing the target variable that the model should predict

        treatment_col (string): the name of the column representing treatment or group that should be considered
            during training

        feature_nums (1d array-like): an optional parameter specifying the number of features to use for
            training. If provided, the model will be trained using only this number of features

    Returns: dict
        A dictionary containing lists of ModelResults for each model trained, indexed by their corresponding
        feature sets
    """
    fast_fit_result = dict()

    for key in features.keys():
        if feature_nums is None:
            feature_nums = [len(features[key])]

        result_n = list()

        for n in feature_nums:
            assert n <= len(
                features[key]
            ), f"Число признаков {n} в feature_nums, не может быть больше числа признаков {len(features[key])}"

            for uplift_type in ["propensity", "abs", "rel"]:

                if uplift_type == "rel" and type(estimator).__name__ in [
                    "CatBoostClassifier",
                    "UpliftRandomForestClassifier",
                    "AufXLearner",
                ]:
                    break

                if (
                    uplift_type == "propensity"
                    and type(estimator).__name__ != "CatBoostClassifier"
                ):
                    continue

                result_n.append(
                    fit_model(
                        estimator,
                        df_train,
                        features[key][:n],
                        target_col,
                        treatment_col,
                        uplift_type,
                    )
                )

        fast_fit_result[key] = result_n

    return fast_fit_result


def get_default_model(model_class: Pipeline):
    """Get default model for each class"""
    import copy

    if model_class == SoloModel:
        return copy.deepcopy(SOLO_MODEL)
    if model_class == TwoModels:
        return copy.deepcopy(TWO_MODEL)
    if model_class == AufTreeClassifier:
        return copy.deepcopy(UPLIFT_TREE)
    if model_class == AufRandomForestClassifier:
        return copy.deepcopy(UPLIFT_FOREST)
    if model_class == AufXLearner:
        return copy.deepcopy(X_LEARNER)
    raise ValueError("Unexpected value of model_class")


def get_default_params_dict(model_class: Pipeline):
    """Get default model for each class"""
    if model_class == SoloModel:
        return SOLO_MODEL.get_params()
    if model_class == TwoModels:
        return TWO_MODEL.get_params()
    if model_class == AufTreeClassifier:
        return UPLIFT_TREE.get_params()
    if model_class == AufRandomForestClassifier:
        return UPLIFT_FOREST.get_params()
    if model_class == AufXLearner:
        return X_LEARNER.get_params()
    raise ValueError("Unexpected value of model_class")



import warnings

import numpy as np
import pandas as pd
from catboost import CatBoostClassifier
from causalml.inference.tree import UpliftRandomForestClassifier

warnings.filterwarnings("ignore")


class AufRandomForestClassifier(UpliftRandomForestClassifier):
    def __init__(
        self,
        control_name,
        n_estimators=10,
        max_features=10,
        random_state=None,
        max_depth=5,
        min_samples_leaf=100,
        min_samples_treatment=10,
        n_reg=10,
        early_stopping_eval_diff_scale=1,
        evaluationFunction="KL",
        normalization=True,
        honesty=False,
        estimation_sample_size=0.5,
        n_jobs=-1,
        joblib_prefer: str = "threads",
    ):
        self.params = {
            "control_name": control_name,
            "n_estimators": n_estimators,
            "max_features": max_features,
            "random_state": random_state,
            "max_depth": max_depth,
            "min_samples_leaf": min_samples_leaf,
            "min_samples_treatment": min_samples_treatment,
            "n_reg": n_reg,
            "early_stopping_eval_diff_scale": early_stopping_eval_diff_scale,
            "evaluationFunction": evaluationFunction,
            "normalization": normalization,
            "honesty": honesty,
            "estimation_sample_size": estimation_sample_size,
            "n_jobs": n_jobs,
            "joblib_prefer": joblib_prefer,
        }
        super().__init__(**self.params)
        self.base_model = CatBoostClassifier(
            n_estimators=200, learning_rate=0.05, max_depth=4, silent=True
        )
        self.ctrl_preds_ = None
        self.trmnt_preds_ = None

    def get_params(self):
        # Вызываем метод get_params из родительского класса и добавляем новый параметр
        return self.params

    def fit(self, X: pd.DataFrame, y: pd.Series, treatment: pd.Series):
        self.feature_names_ = X.columns.tolist()
        fltr = treatment.astype("str") == self.params["control_name"]
        self.base_model.fit(X=X[fltr], y=y[fltr])
        with np.errstate(divide="ignore", invalid="ignore"):
            super().fit(
                X=X.values, y=y.values, treatment=treatment.astype("str").values
            )
        return super()

    def predict(self, X: pd.DataFrame):
        self.ctrl_preds_ = self.base_model.predict_proba(X)[:, 1]
        with np.errstate(divide="ignore", invalid="ignore"):
            uplift = super().predict(X.values).reshape(-1)
        self.trmnt_preds_ = self.ctrl_preds_ + uplift
        return uplift



from .binary import UpliftPipeline
from .calibration import UpliftCalibration
from .inference import UpliftPipelineInference

__all__ = [
    "UpliftPipeline",
    "UpliftCalibration",
    "UpliftPipelineInference"
]


import typing as tp

import numpy as np
import pandas as pd
from catboost import CatBoostClassifier
from causalml.feature_selection import FilterSelect
from causalml.inference.tree import UpliftRandomForestClassifier
from sklift.models import SoloModel


class ImportanceRanker:
    """Sorts features by their importance scores from some model.

    Uses one of available uplift or propensity modelling methods which provides feature_importances_
    Sorts features either at one time or recursively elliminating features by one or some fixed percent
    This ranker can work only with train dataset to calculate feature importances

    Note: 'dropping' below means "don't consider it in further ranking process", not "remove feature from the feature list"

    Args:
        model_class ([propensity, s-learner, uplift tree]): uplift model object
        model_params (dict): dictionary of model parameters
        sorting_mode (string, ['at_once', 'iterative', 'rfe']): features sorting method to be used
            'at_once' for just sorting features with single model
            'iterative' for 'dropping' 10 % worst feature (by importance scores) each time till 1 feature remains
            'rfe' for 'dropping' one worst feature (by importance scores) each time till 1 feature remains
        name (string): ranker name
        data (pd.DataFrame): DataFrame containing target, features, and treatment group
        features (1d array-like): list of feature names, that are columns in the data DataFrame
        target_col (string): target column name
        treatment_col (string): treatment column name

    Examples:
        >>> from AUF.feature_selection import ImportanceRanker
        >>> ranker = ImportanceRanker(SoloModel, model_params, "at_once")
        >>> ranked_features, ranked_importances = ranker.run(df, features, target_col, treatment_col)

    Returns: None
    """

    def __init__(
        self,
        model_class: tp.Any,
        model_params: tp.Dict[str, tp.Any],
        sorting_mode: str,
        name: str = "__empty_name__",
    ):

        assert sorting_mode in ("at_once", "iterative", "rfe")

        try:
            try_to_create_model = model_class(**model_params)
            assert "fit" in list(dir(try_to_create_model))
            assert (
                set(["predict", "predict_proba"]) & set(dir(try_to_create_model))
                != set()
            )
            del try_to_create_model
        except Exception as exc:
            raise ValueError(
                "Check params for creating model: model creation failed"
            ) from exc

        self._model_class = model_class
        self._model_params = model_params
        self._sorting_mode = sorting_mode
        self._name: str = name
        self._ranked_features: tp.List[str] = []
        self._ranked_features_scores: tp.List[float] = []

    def _run_propensity(
        self, data: pd.DataFrame, features: tp.List[str], target_col: str
    ) -> tp.List[str]:
        model = self._model_class(**self._model_params)
        model.fit(data[features], data[target_col])
        importances = model.feature_importances_
        features_info = list(zip(importances, features))
        features_info = sorted(features_info, key=lambda x: -x[0])
        importances, ranked_features = map(list, zip(*features_info))
        return ranked_features, importances

    def _run_s_learner(
        self,
        data: pd.DataFrame,
        features: tp.List[str],
        target_col: str,
        treatment_col: str,
    ) -> tp.List[str]:
        model = self._model_class(**self._model_params)
        model.fit(data[features], data[target_col], data[treatment_col])
        importances = model.estimator.feature_importances_
        feature_names = model.estimator.feature_names_
        features_info = list(zip(importances, feature_names))
        features_info = [
            p for p in features_info if p[1] not in [treatment_col, "treatment"]
        ]
        features_info = sorted(features_info, key=lambda x: -x[0])
        importances, ranked_features = map(list, zip(*features_info))
        return ranked_features, importances

    def _run_uplift_forest(
        self,
        data: pd.DataFrame,
        features: tp.List[str],
        target_col: str,
        treatment_col: str,
    ) -> tp.List[str]:
        model = self._model_class(**self._model_params)
        model.fit(
            X=data[features].values,
            treatment=data[treatment_col]
            .apply(lambda t: "treatment" if t == 1 else "control")
            .values,
            y=data[target_col].values,
        )
        importances = model.feature_importances_
        features_info = list(zip(importances, features))
        features_info = sorted(features_info, key=lambda x: -x[0])
        importances, ranked_features = map(list, zip(*features_info))
        return ranked_features, importances

    def _run(
        self,
        data: pd.DataFrame,
        features: tp.List[str],
        target_col: str,
        treatment_col: str,
    ) -> tp.List[str]:
        """Just calls needed model runner"""

        # self.model_info = (class, rfe-mode, params dict)
        ranked_features: tp.List[str] = []
        importances: tp.List[float] = []

        if self._model_class is SoloModel:
            ranked_features, importances = self._run_s_learner(
                data, features, target_col, treatment_col
            )
        elif self._model_class is UpliftRandomForestClassifier:
            ranked_features, importances = self._run_uplift_forest(
                data, features, target_col, treatment_col
            )
        else:
            # rank features with propensity model
            ranked_features, importances = self._run_propensity(
                data, features, target_col
            )

        return ranked_features, importances

    def run(
        self,
        data: pd.DataFrame,
        features: tp.List[str],
        target_col: str,
        treatment_col: str,
    ) -> tp.Tuple[tp.List[str], tp.List[float]]:
        """Defines the strategy of ranking features (at once, iterative, rfe)"""

        # self.model_info = (class, rfe-mode, params dict)
        ranked_features: tp.List[str] = features.copy()
        importances: tp.List[float] = [0.0 for f in ranked_features]

        if self._sorting_mode == "at_once":
            ranked_features, importances = self._run(
                data, ranked_features, target_col, treatment_col
            )
        elif self._sorting_mode == "iterative":
            # 10%
            n_top = len(features)
            while n_top:
                feats, imps = self._run(
                    data, ranked_features[:n_top], target_col, treatment_col
                )
                ranked_features[:n_top] = feats
                importances[:n_top] = imps
                n_top = int(n_top * 0.9)
        else:
            # rfe
            for n_top in range(len(features), 0, -1):
                feats, imps = self._run(
                    data, ranked_features[:n_top], target_col, treatment_col
                )
                ranked_features[:n_top] = feats
                importances[:n_top] = imps

        self._ranked_features = ranked_features.copy()
        self._ranked_features_scores = importances.copy()
        return self._ranked_features, self._ranked_features_scores

    def get_ranker_name(self) -> str:
        return self._name

    def get_ranked_features(self) -> tp.List[str]:
        return self._ranked_features

    def get_ranked_features_scores(self) -> tp.List[float]:
        return self._ranked_features_scores



from .plots import (
    plot_uplift_by_feature_bins,
    plot_portrait_tree,
    plot_uplift_by_percentile
)

__all__ = [
    "plot_uplift_by_feature_bins",
    "plot_portrait_tree",
    "plot_uplift_by_percentile"
]


import typing as tp

import numpy as np
import pandas as pd
from sklearn.base import BaseEstimator


class AufXLearner:
    """X-learner с поддержкой eval_set

    Обучает две различные response модели для группы с взаимодействием и без него
    Затем строим прокси-модели по предсказанию
        uplift = (y_true - control_model  | treatment = 1)
        uplift = (treatment_model- y_true | treatment = 0)

    Итоговый аплифт uplift_final = eps * uplift_control_model + (1-eps)*uplift_treatment_model

    Больше о работе AufXLearner:
        https://habr.com/ru/companies/glowbyte/articles/686398/

    Args:
        model (estimator object implementing 'fit'):
            The object to use to fit the treatment data.
        uplift_model (estimator object implementing 'fit'):
            The object to use to fit the control data.
        map_groups (dict):
            словарь "группа из данных - группа для Xlearner"
            обязательно должен присутствовать ключ control
        features (list):
            фичи которые будут использовать модели
        cat_features (list):
            категориальные фичи которые будут использовать модели
        group_model (estimator object implementing 'fit'):
            The object to use to fit the control data.

    Attributes:
        trmnt_preds_ (array-like, shape (n_samples, )):
            Estimator predictions uplift by uplift_treatment_model
        ctrl_preds_ (array-like, shape (n_samples, )):
            Estimator predictions uplift by uplift_control_model

    Example:

        from rb_ds_tools.uplift.uplift_models import XLearnerWeight
        from catboost import CatBoostClassifier, CatBoostRegressor

        params = {
            "iterations": 500,
            "depth": 8,
            "learning_rate": 0.1,
            "loss_function": 'Logloss',
            'early_stopping_rounds': 50,
            'eval_metric': 'AUC',
            'random_seed': RANDOM_SEED,
        }

        params_reg = {
            "iterations": 300,
            "depth": 8,
            "learning_rate": 0.01,
            "loss_function": 'MAE',
            'nan_mode': 'Min',
            'eval_metric': 'MAE',
            'use_best_model': True,
            'verbose': False,
            'random_state': RANDOM_SEED,
        }


        clf = XLearnerGroup(
            model=CatBoostClassifier(**params),
            uplift_model=CatBoostRegressor(**params_reg),
            map_groups = {'control': 0, 'treatment1': 1},
            features=int_cols + cat_cols,
            cat_features=cat_cols,
            group_model = CatBoostClassifier(**params)
        )

        '''
            Если у нас есть готовые response модели для каждой из групп

            treat_model = load_model(path1)
            cntrl_model = load_model(path2)
            response_features = list(set(treat_model.feature_names_) + set(treat_model.feature_names_))
            models = {'control': treat_model, 'treatment1':cntrl_model}
            clf.add_fit_response(models=models, features=response_features)
        '''

        clf.fit(
            pd.concat([x_train, y_train], axis = 1),
            pd.concat([x_val, y_val], axis = 1),
            treat_col='treatment',
            target_col='target',
            cost_dict={
                'control': 1,
                'treatment1': 1
            }
        )

        uplift = clf.predict(x_test)['treatment1']

    """

    def __init__(
        self,
        model,
        uplift_model,
        map_groups={"control": 0, "treatment1": 1, "treatment2": 2},
        features=None,
        cat_features=None,
        group_model=None,
    ):
        self._model = {}
        self._uplift_model = {}
        self._params = {
            "model": model.copy(),
            "uplift_model": uplift_model.copy(),
        }

        self._map_groups = map_groups
        self._features = features
        self._response_features = features
        self._cat_features = cat_features
        self._feature_importances = None
        self._group_model = group_model

        assert (
            "control" in self._map_groups.keys()
        ), "map_groups должен содержать ключ control"

        for key in self._map_groups.keys():
            self._model[key] = model.copy()
            self._uplift_model[key] = {
                "control": uplift_model.copy(),
                "treatment": uplift_model.copy(),
            }

    def get_params(self):
        return self._params

    def get_feature_importances(self):
        if self._feature_importances is None:
            return self._get_feature_importances()
        return self._feature_importances

    def _get_feature_importances(self):
        self._feature_importances = {}

        for key in self._map_groups.keys():
            if key != "control":
                fi_cntrl = self._uplift_model[key]["control"].get_feature_importance(
                    prettified=True
                )
                fi_treat = self._uplift_model[key]["treatment"].get_feature_importance(
                    prettified=True
                )

                fi = pd.merge(fi_cntrl, fi_treat, on="Feature Id", how="inner")
                fi["Importances"] = fi.Importances_x + fi.Importances_y
                fi.Importances = fi.Importances.apply(lambda x: round(x, 3))
                fi.index = fi["Feature Id"]
                fi = fi.drop(
                    ["Feature Id", "Importances_x", "Importances_y"], axis=1
                ).to_dict()["Importances"]

                self._feature_importances[key] = fi

        return self._feature_importances

    def fit(
        self,
        X: pd.DataFrame,
        y: pd.Series,
        treatment: pd.Series,
        cost_dict: tp.Optional[tp.Dict[tp.Any, float]] = None,
        **qwargs,
    ):
        """Обучение модели

        Args:
            X:
                должен содержать фичи из self._features и self._response_features
            y:
                колонка с таргетом
            treatment:
                колонка с информацией о группе клиента
            cost_dict:
                доходность утилизаций для разных групп

        Returns:
            object: self
        """

        if cost_dict is None:
            cost_dict = {}
            for group in self._map_groups:
                cost_dict[group] = 1

        assert (
            len(set(self._map_groups.keys()).symmetric_difference(cost_dict.keys()))
            == 0
        ), f"Ключи cost_dict:{cost_dict.keys()} не совпадают с ключами групп map_groups:{self._map_groups.keys()}"

        if self._features is None:
            self._features = X.columns.tolist()

        if self._cat_features is None:
            self._cat_features = [f for f in self._features if X[f].dtype == "object"]

        if self._response_features is None:
            self._response_features = X.columns.tolist()

        assert set(X.columns) == set(self._features) | set(
            self._response_features
        ), "X должен содержать все фичи из self._features и self._response_features"

        y = np.array(y)
        treatment = np.array(treatment)

        if cost_dict is None:
            for key in self._map_groups.keys():
                cost_dict[key] = 1

        if self._group_model is not None:
            self._group_model = self._group_model.fit(
                X=X.loc[:, self._features],
                y=treatment,
                cat_features=self._cat_features,
                **qwargs,
            )

        # обучили все response модели
        for key in self._map_groups.keys():
            treat_name = self._map_groups[key]
            mask = treatment == treat_name
            self._model[key] = self._model[key].fit(
                X=X.loc[mask, self._features],
                y=y[mask],
                cat_features=self._cat_features,
                **qwargs,
            )

        cntrl_name = self._map_groups["control"]
        cntrl_mask = treatment == cntrl_name
        cntrl_X_resp = X.loc[cntrl_mask, self._response_features]
        cntrl_X = X.loc[cntrl_mask, self._features]

        for key in self._map_groups.keys():
            if key != "control":
                trmnt_name = self._map_groups[key]

                sj = cost_dict[key]
                s0 = cost_dict["control"]

                # fit uplift model for control group
                cntrl_probas = self._model[key].predict_proba(cntrl_X_resp)[:, 1]
                cntrl_uplift = sj * cntrl_probas - s0 * y[cntrl_mask]

                self._uplift_model[key]["control"] = self._uplift_model[key][
                    "control"
                ].fit(
                    X=cntrl_X, y=cntrl_uplift, cat_features=self._cat_features, **qwargs
                )

                # fit uplift model for treatment group
                trmnt_mask = treatment == trmnt_name
                trmnt_X_resp = X.loc[trmnt_mask, self._response_features]
                trmnt_X = X.loc[trmnt_mask, self._features]

                trmnt_probas = self._model["control"].predict_proba(trmnt_X_resp)[:, 1]
                trmnt_uplift = -s0 * trmnt_probas + sj * y[trmnt_mask]

                self._uplift_model[key]["treatment"] = self._uplift_model[key][
                    "treatment"
                ].fit(
                    X=trmnt_X, y=trmnt_uplift, cat_features=self._cat_features, **qwargs
                )

        return self

    def predict(self, X: pd.DataFrame, eps: float = 0.5):
        """Get Xlearner uplift prediction

        Args:
            X:
                sample for uplift prediction
            eps:
                is used only if self._group_model is not None
                weight for control uplift model predictions
                (1 - eps) = weight for treatment uplift model predictions

        Returns:
            dict of arrays:
                uplift for different treatment groups
        """

        uplifts = {}

        for key in self._map_groups.keys():
            if key != "control":

                if self._group_model is not None:
                    eps = self._group_model.predict_proba(X[self._features])[:, 0]

                t0 = self._uplift_model[key]["control"].predict(X[self._features])
                t1 = self._uplift_model[key]["treatment"].predict(X[self._features])
                uplifts[key] = eps * t0 + (1 - eps) * t1

        if len(uplifts) == 1:
            treatment_group_name = [group for group in uplifts if group != "control"][0]
            assert (
                len(uplifts[treatment_group_name]) == X.shape[0]
            ), f"len(uplifts[treatment_group_name]) = {len(uplifts[treatment_group_name])}, X.shape[0] = {X.shape[0]}"
            return uplifts[treatment_group_name]

        return uplifts



import warnings

import numpy as np
import pandas as pd
from sklearn.base import BaseEstimator
from sklearn.utils.multiclass import type_of_target
from sklearn.utils.validation import check_consistent_length
from sklift.utils import check_is_binary

"""S-learner with mulltitreatment

1) multitreatment = treatment flag as a categorical feature
2) uplift prediction = uplift for every treatment
3) is treatment interaction approach possible?
        {feats} -> {feats_T_1} + {feats_T_2} + ...
        {feats_T_1} = {feature * T_1_flag for feature in feats}
        10 treatments --> 2_000 feats = 20_000 feats...
4) trmnt_preds_ & ctrl_preds_ == ? dict[treat, proba]?
5) 
"""


class AufSoloModel(BaseEstimator):
    """aka Treatment Dummy approach, or Single model approach, or S-Learner.

    Fit solo model on whole dataset with 'treatment' as an additional feature.

    Each object from the test sample is scored twice: with the communication flag equal to 1 and equal to 0.
    Subtracting the probabilities for each observation, we get the uplift.

    Return delta of predictions for each example.

    Read more in the :ref:`User Guide <SoloModel>`.

    Args:
        estimator (estimator object implementing 'fit'): The object to use to fit the data.
        method (string, ’dummy’ or ’treatment_interaction’, default='dummy'): Specifies the approach:

            * ``'dummy'``:
                Single model;
            * ``'treatment_interaction'``:
                Single model including treatment interactions.

    Attributes:
        trmnt_preds_ (array-like, shape (n_samples, )): Estimator predictions on samples when treatment.
        ctrl_preds_ (array-like, shape (n_samples, )): Estimator predictions on samples when control.

    Example::

        # import approach
        from auf.models import AufSoloModel
        # import any estimator adheres to scikit-learn conventions
        from catboost import CatBoostClassifier


        asm = AufSoloModel(CatBoostClassifier(verbose=100, random_state=777))  # define approach
        asm = asm.fit(X_train, y_train, treat_train, estimator_fit_params={{'plot': True})  # fit the model
        uplift_asm = asm.predict(X_val)  # predict uplift

    References (binary treatment):
        Lo, Victor. (2002). The True Lift Model - A Novel Data Mining Approach to Response Modeling
        in Database Marketing. SIGKDD Explorations. 4. 78-86.

    See Also:

        **Other approaches:**

        * :class:`.ClassTransformation`: Class Variable Transformation approach.
        * :class:`.ClassTransformationReg`: Transformed Outcome approach.
        * :class:`.TwoModels`: Double classifier approach.

        **Other:**

        * :func:`.plot_uplift_preds`: Plot histograms of treatment, control and uplift predictions.
    """

    def __init__(self, estimator, method="dummy"):
        self.estimator = estimator
        self.method = method
        self.trmnt_preds_ = None
        self.ctrl_preds_ = None
        self._type_of_target = None

        self._treatment_values = None
        self._control_group = None

        all_methods = ["dummy", "treatment_interaction"]
        if method not in all_methods:
            raise ValueError(
                "SoloModel approach supports only methods in %s, got"
                " %s." % (all_methods, method)
            )

    def _preprocess_data(self, X, treatment):
        if self.method == "dummy":
            if isinstance(X, np.ndarray):
                X_mod = np.column_stack((X, treatment))
            elif isinstance(X, pd.DataFrame):
                X_mod = X.assign(treatment=treatment)
            else:
                raise TypeError(
                    "Expected numpy.ndarray or pandas.DataFrame in training vector X, got %s"
                    % type(X)
                )

        if self.method == "treatment_interaction":
            # TODO : add features * indicators(treatment == T_i)
            X_mod = X.copy()

            for treatment_id in sorted(self._treatment_values):
                if treatment_id == self._control_group:
                    continue

                if isinstance(X_mod, np.ndarray):
                    X_mod = np.column_stack(
                        (
                            X_mod,
                            np.multiply(
                                X, np.array(treatment == treatment_id).reshape(-1, 1)
                            ),
                            # treatment
                        )
                    )
                elif isinstance(X_mod, pd.DataFrame):
                    X_mod = pd.concat(
                        [
                            X_mod,
                            X.apply(lambda x: x * (treatment == treatment_id)).rename(
                                columns=lambda x: str(x)
                                + f"_{treatment_id}"
                                + "_treatment_interaction"
                            ),
                        ],
                        axis=1,
                    )
                    # .assign(treatment=treatment)
                else:
                    raise TypeError(
                        "Expected numpy.ndarray or pandas.DataFrame in training vector X, got %s"
                        % type(X)
                    )

            if isinstance(X_mod, np.ndarray):
                X_mod = np.column_stack((X_mod, treatment))
            elif isinstance(X_mod, pd.DataFrame):
                X_mod = X_mod.assign(treatment=treatment)

        return X_mod

    def fit(self, X, y, treatment, control_group, estimator_fit_params=None):
        """Fit the model according to the given training data.

        For each test example calculate predictions on new set twice: by the first and second models.
        After that calculate uplift as a delta between these predictions.

        Return delta of predictions for each example.

        Args:
            X (array-like, shape (n_samples, n_features)): Training vector, where n_samples is the number of
                samples and n_features is the number of features.
            y (array-like, shape (n_samples,)): Binary target vector relative to X.
            treatment (array-like, shape (n_samples,)): Binary treatment vector relative to X.
            estimator_fit_params (dict, optional): Parameters to pass to the fit method of the estimator.

        Returns:
            object: self
        """

        check_consistent_length(X, y, treatment)
        # check_is_binary(treatment)
        treatment_values = np.unique(treatment)

        self._treatment_values = treatment_values.copy()
        self._control_group = control_group

        # now we expect any treatment groups number except 1
        # if len(treatment_values) != 2:
        #     raise ValueError("Expected only two unique values in treatment vector, got %s" % len(treatment_values))
        if len(treatment_values) == 1:
            raise ValueError("Expected more than one unique values in treatment vector")

        X_mod = self._preprocess_data(X, treatment)

        self._type_of_target = type_of_target(y)

        if estimator_fit_params is None:
            estimator_fit_params = {}
        self.estimator.fit(X_mod, y, **estimator_fit_params)
        return self

    def predict(self, X):
        """Perform uplift on samples in X.

        Args:
            X (array-like, shape (n_samples, n_features)): Training vector, where n_samples is the number of samples
                and n_features is the number of features.

        Returns:
            array (shape (n_samples,)): uplift
        """

        self._preds = {}

        for treatment_id in sorted(self._treatment_values):
            treatment = np.full(X.shape[0], treatment_id)
            X_mod = self._preprocess_data(X, treatment)

            if self._type_of_target == "binary":
                self._preds[treatment_id] = self.estimator.predict_proba(X_mod)[:, 1]
            else:
                self._preds[treatment_id] = self.estimator.predict(X_mod)

        uplifts = {}

        for treatment_id in sorted(self._treatment_values):
            if treatment_id == self._control_group:
                continue
            uplifts[treatment_id] = (
                self._preds[treatment_id] - self._preds[self._control_group]
            )

        return uplifts



import typing as tp

import numpy as np
import pandas as pd
import scipy.stats as ss
from catboost import CatBoostClassifier
from sklearn.metrics import roc_auc_score

from ..constants import BOOTSTRAP_REPEATS, RANDOM_STATE


def check_bernoulli_dependence(x: np.array, y: np.array, alpha: float = 0.05):
    """Check if 2 binary flags are dependent

    Args:
        x, y (1d array-like): paired observations of flags
        alpha (float): max level for error of the 1-st type. Default 0.05

    Returns:
        dict: Result of dependency check
        {'stat': stat_value,
         'critical_stat': critical_stat_value,
         'dependent': bool_decision (1 == dependent == stat > critical),
         'pvalue': pvalue
        }
    """
    assert len(x) == len(y), f"{len(x)} != {len(y)}"
    assert set(np.unique(x)) == set([0, 1]), f"set(np.unique(x)) = {set(np.unique(x))}"
    assert set(np.unique(y)) == set([0, 1]), f"set(np.unique(y)) = {set(np.unique(y))}"
    assert 0.0 < alpha < 1.0

    n = len(x)
    co_matr = np.array([[0, 0], [0, 0]])
    for i in [0, 1]:
        for j in [0, 1]:
            co_matr[i, j] = np.sum((x == i) * (y == j))

    stat_sum = 0.0
    for i in [0, 1]:
        for j in [0, 1]:
            x_i = np.sum(co_matr[i, :])
            y_j = np.sum(co_matr[:, j])
            stat_sum += co_matr[i, j] ** 2 / (x_i * y_j)

    stat = n * (stat_sum - 1)

    critical_stat = ss.chi2(df=1).ppf(1 - alpha)
    pvalue = 1 - ss.chi2(df=1).cdf(stat)

    return {
        "stat": stat,
        "critical_stat": critical_stat,
        "dependent": stat > critical_stat,
        "pvalue": pvalue,
    }


def check_bernoulli_equal_means(x: np.array, y: np.array, alpha: float = 0.05):
    """Check if 2 binary flags have the same mean

    Args:
        x, y (1d array-like): (generally not paired) observations of flags
        alpha (float): max level for error of the 1-st type

    Returns:
        dict: Result of equal means check
        {'stat': stat_value,
         # 'critical_stat': critical_stat_value,
         'equals': bool_decision (1 == dependent == stat > critical),
         'pvalue': pvalue
         }
    """
    stat, pvalue = ss.ttest_ind(
        x,
        y,
        equal_var=False,
        nan_policy="propagate",
        alternative="two-sided",
    )

    return {
        "stat": stat,
        # 'critical_stat': critical_stat_value,
        "equals": pvalue >= alpha,
        "pvalue": pvalue,
    }


def check_nans(
    df: pd.DataFrame, feature_cols: tp.List[str], max_nan_ratio: float = 0.95
):
    """Check if the percentage of nans in a feature exceeds a certain specified threshold
       and then filter out the features where it does

    Args:
        df (pd.DataFrame): dataframe with features
        feature_cols (1d array-like): features to check
        max_nan_ratio (float): maximum allowable nans ratio for each feature
            in all (train / val / test) sets together

    Returns:
        list: Column names in which the percentage of nans does not exceed the specified threshold
    """
    return [f for f in feature_cols if df[f].isna().mean() <= max_nan_ratio]


# TO-DO: переименовать функцию в check_not_enough_unique_values или check_few_unique_values или вообще check_unique_values
def check_too_less_unique_value(df: pd.DataFrame, feature_cols: tp.List[str]):
    """Check that features have at least 2 unique values, filter out others

    Args:
        df (pd.DataFrame): dataframe with features
        feature_cols (1d array-like): features to check

    Returns:
        list: Column names in which the number of unique values is greater than one
    """
    return [f for f in feature_cols if df[f].nunique() >= 2]


# TO-DO: переименовать функцию в check_not_enough_unique_values или check_few_unique_values или вообще check_unique_category_values
def process_too_much_categories(
    df: pd.DataFrame, feature_cols: tp.List[str], max_categories_count: int = 20
):
    """Check that categorical features do not have more than specified number of unique values,
        if they do, leave the top of categories, all rest are replaced by "_others_"

    Args:
        df (pd.DataFrame): dataframe with features
        feature_cols (1d array-like): features to check
        max_categories_count (float): maximum allowable number of unique values for categorical
            feature in all (train / val / test) sets together

    Returns: None
    """
    for col in feature_cols:
        if df[col].dtype == "object" and df[col].nunique() >= max_categories_count:
            sorted_values_desc = df[col].value_counts().sort_values()[::-1]
            others = sorted_values_desc[max_categories_count - 1 :].index.tolist()
            df.loc[df[col].isin(others), col] = "_others_"


# TO-DO: Нужно ли вместе с early_stopping добавить допустимое уменьшение roc auc?
def check_leaks_v2(
    df: pd.DataFrame,
    base_cols_mapper: tp.Dict[str, str],
    feature_cols: tp.List[str],
    col_to_check: str,
    alpha: float = 0.05,
    max_val_roc_auc: float = 0.6,
    early_stopping: int = None,
):
    """Find features with leak on target or treatment using simple ML model

    Args:
        df (pd.DataFrame): dataframe with features
        base_cols_mapper (dict): dictionary of correspondences between the columns names used
            in library functions and names in input data
        feature_cols (1d array-like): features to check
        col_to_check (string): 'target' or 'treatment' -- names of columns to check for leaks
        alpha (float): significance level
        max_val_roc_auc (float): maximum allowable ROC-AUC score for predicting col_to_check
            by any single feature on validation set
        early_stopping (int): number of consecutive stages dirung which the model's preformance
            on the validation set plateaus or worsens

    One of preselection steps which helps to search appropriate feature candidates for modelling

    Idea:
        - build a small model for check_to_col prediction
        - assess degree of the total leak from all features with procedure which is described below
        - if leak is statistically significant, then drop top-1 feature (with the greatest feature_importance_) and repeat
        - else stop algorithm and return all other "clean" features

    Leak assession procedure description:
        - small model (CatBoostClassifier) is fitted on training set using all available features
        - probability predictions are made for validation set
        - these probabilities are bootstrapped 500 times together with corresponding targets
        - ROC-AUC is calculated for each bootstrapped probabilities and targets sample
        - all ROC-AUC scores less than 0.5 are transformed with function f(x) = 1 - x (x := max(x, 1-x))
        - (1 - alpha) quantile is computed for these 500 scores
        - if estimated quantile is less than max_val_roc_auc then leak is thought to be not such significant
        - if estimated quantile is more than max_val_roc_auc then leak is thought to exist

     Advantage over simple filter used in original function:
        - this method consideres features interactions in treatment prediction
        - for example, 3 features each with ROC-AUC 0.6 may give ROC-AUC 0.8 used together in one model

    Returns:
        leaks_roc_aucs (list): Feature names and ROC-AUCs of leaking features
        not_leaks (list): Feature names of non leaking features
        all_features_roc_aucs (list): Feature name and ROC-AUC for each feature
    """
    assert col_to_check in [
        "target",
        "treatment",
    ], f"Error : col_to_check must be in ['target', 'treatment'] but col_to_check='{col_to_check}'"
    assert (
        0.0 < alpha and alpha < 1.0
    ), f"alpha must be in [0.0, 1.0] but is {alpha:.2f}"
    assert (
        max_val_roc_auc > 0.5
    ), f"max_val_roc_auc must be greater than 0.5, but is {max_val_roc_auc:.2f}"

    features = feature_cols.copy()
    all_features_roc_aucs = []

    segm_col = base_cols_mapper["segm"]
    col_to_check = base_cols_mapper[col_to_check]

    while len(features) > 0:
        feature_df_train = df.loc[
            df[segm_col] == "train", features + [col_to_check]
        ].copy()
        feature_df_val = df.loc[df[segm_col] == "val", features + [col_to_check]].copy()

        bootstrap_data_size = max(
            15_000, min(feature_df_val.shape[0], int(df.shape[0] * 0.1))
        )
        feature_df_val = feature_df_val.sample(bootstrap_data_size, random_state=8)

        for f in features:
            if df[f].dtype == "object":
                feature_df_train[f] = feature_df_train[f].astype(str)
                feature_df_val[f] = feature_df_val[f].astype(str)

        checker = CatBoostClassifier(
            n_estimators=10,
            depth=2,
            learning_rate=0.1,
            silent=True,
            random_seed=8,
            cat_features=[f for f in features if df[f].dtype == "object"],
        )

        checker.fit(feature_df_train[features], feature_df_train[col_to_check])
        val_preds = checker.predict_proba(feature_df_val[features])[:, 1]
        val_true = feature_df_val[col_to_check].values

        bootstrap_roc_aucs = np.zeros(shape=BOOTSTRAP_REPEATS)
        rng = np.random.RandomState(seed=RANDOM_STATE)

        for boot in range(BOOTSTRAP_REPEATS):
            idxs = rng.choice(range(len(val_true)), size=len(val_true), replace=True)
            true, preds = val_true[idxs], val_preds[idxs]
            roc_auc = roc_auc_score(true, preds)
            bootstrap_roc_aucs[boot] = roc_auc

        # other (simpler?) but equivalent way to determine leak ?
        # pvalue = np.mean(bootstrap_roc_aucs >= max_val_roc_auc)
        # if pvalue >= alpha:
        #     not_leaks.append(f)

        q = np.quantile(bootstrap_roc_aucs, q=1 - alpha)
        q = max(q, 1 - q)  # revert labels if needed

        # if q < max_val_roc_auc:
        #     # this max allowed value is quite rare
        #     # all leaks found
        #     break

        top_leaking_feature = features[np.argmax(checker.feature_importances_)]
        features.remove(top_leaking_feature)
        all_features_roc_aucs.append((top_leaking_feature, q))

        if early_stopping is not None:
            if early_stopping == 0:
                break
            if q < max_val_roc_auc:
                early_stopping -= 1

    leaks_roc_aucs = [(f, q) for f, q in all_features_roc_aucs if q >= max_val_roc_auc]
    not_leaks = [
        f for f, q in all_features_roc_aucs if q < max_val_roc_auc
    ] + features  # not saved as leaks --> not leaks

    leaks_roc_aucs = sorted(leaks_roc_aucs, key=lambda p: -p[1])
    all_features_roc_aucs = sorted(all_features_roc_aucs, key=lambda p: -p[1])

    return leaks_roc_aucs, not_leaks, all_features_roc_aucs


def check_leaks(
    df: pd.DataFrame,
    base_cols_mapper: tp.Dict[str, str],
    feature_cols: tp.List[str],
    col_to_check: str,
    alpha: float = 0.05,
    max_val_roc_auc: float = 0.6,
):
    """Find features with leak on target or treatment using simple ML model

    Args:
        df (pd.DataFrame): dataframe with features
        base_cols_mapper (dict): dictionary of correspondences between the columns names used
            in library functions and names in input data
        feature_cols (1d array-like): features to check
        col_to_check (string): 'target' or 'treatment' -- names of columns to check for leaks
        alpha (float): significance level
        max_val_roc_auc (float): maximum allowable ROC-AUC score for predicting col_to_check
            by any single feature on validation set

    One of preselection steps which helps to search appropriate feature candidates for modelling

    Idea: (1 - alpha) quantile of validation ROC-AUC score in col_to_check prediction is estimated
        for each single feature independently with following procedure:
        - small model (CatBoostClassifier) is fitted on training set using only this one feature
        - probability predictions are made for validation set
        - these probabilities are bootstrapped 500 times together with corresponding targets
        - ROC-AUC is calculated for each bootstrapped probabilities and targets sample
        - all ROC-AUC scores less than 0.5 are transformed with function f(x) = 1 - x
        - compute (1 - alpha) quantile of these 500 scores

    The rule of leak detection for each feature is:
        - if estimated quantile is less than max_val_roc_auc then leak is thought to be not such significant
        - if estimated quantile is more than max_val_roc_auc then leak is thought to exist

    Returns:
        leaks_roc_aucs (list): Feature names and ROC-AUCs of leaking features
        not_leaks (list): Feature names of non leaking features
        all_features_roc_aucs (list): Feature name and ROC-AUC for each feature
    """

    assert col_to_check in [
        "target",
        "treatment",
    ], f"Error : col_to_check must be in ['target', 'treatment'] but col_to_check='{col_to_check}'"
    assert (
        0.0 < alpha and alpha < 1.0
    ), f"alpha must be in [0.0, 1.0] but is {alpha:.2f}"
    assert (
        max_val_roc_auc > 0.5
    ), f"max_val_roc_auc must be greater than 0.5, but is {max_val_roc_auc:.2f}"

    leaks_roc_aucs = []
    not_leaks = []
    all_features_roc_aucs = []

    segm_col = base_cols_mapper["segm"]
    col_to_check = base_cols_mapper[col_to_check]

    for f in feature_cols:
        feature_df_train = df.loc[df[segm_col] == "train", [f, col_to_check]].copy()
        feature_df_val = df.loc[df[segm_col] == "val", [f, col_to_check]].copy()

        if df[f].dtype == "object":
            feature_df_train[f] = feature_df_train[f].astype(str)
            feature_df_val[f] = feature_df_val[f].astype(str)

        checker = CatBoostClassifier(
            n_estimators=30,
            depth=1,
            learning_rate=0.1,
            silent=True,
            random_seed=8,
            cat_features=[f] if df[f].dtype == "object" else [],
        )

        checker.fit(feature_df_train[[f]], feature_df_train[col_to_check])
        val_preds = checker.predict_proba(feature_df_val[[f]])[:, 1]
        val_true = feature_df_val[col_to_check].values

        bootstrap_roc_aucs = np.zeros(shape=BOOTSTRAP_REPEATS)
        rng = np.random.RandomState(seed=RANDOM_STATE)

        for boot in range(BOOTSTRAP_REPEATS):
            idxs = rng.choice(range(len(val_true)), size=len(val_true), replace=True)
            true, preds = val_true[idxs], val_preds[idxs]
            roc_auc = roc_auc_score(true, preds)
            bootstrap_roc_aucs[boot] = roc_auc

        # other (simpler?) but equivalent way to determine leak ?
        # pvalue = np.mean(bootstrap_roc_aucs >= max_val_roc_auc)
        # if pvalue >= alpha:
        #     not_leaks.append(f)

        q = np.quantile(bootstrap_roc_aucs, q=1 - alpha)
        q = max(q, 1 - q)  # revert labels if needed

        all_features_roc_aucs.append((f, q))

        if q < max_val_roc_auc:
            # this max allowed value is quite rare
            not_leaks.append(f)
        else:
            # this value is not so rare
            leaks_roc_aucs.append((f, q))

    leaks_roc_aucs = sorted(leaks_roc_aucs, key=lambda p: -p[1])
    all_features_roc_aucs = sorted(all_features_roc_aucs, key=lambda p: -p[1])

    return leaks_roc_aucs, not_leaks, all_features_roc_aucs


def check_correlations(
    df: pd.DataFrame, feature_cols: tp.List[str], max_abs_corr: float = 0.95
):
    """Check that pairwise correlation of all features does not exeed a certain
        specified threshold, filter out others

    Args:
        df (pd.DataFrame): dataframe with features
        feature_cols (1d array-like): features to check
        max_abs_corr (float): maximum allowable absolute value for correlation between features

    Both train and validation sets are used here to get more precise results

    Algorithm checks for each feature independently:
        - correlations with other features which weren't checked with it yet
        - all features which have correlation greater than max_abs_corr are filtered
        - next feature checks are made with modified feature set

    Returns:
        too_correlated (list): Feature pairs with correlation higher than a given threshold
        features (list): Filtered feature names
    """

    assert (
        0.0 < max_abs_corr and max_abs_corr < 1.0
    ), f"max_abs_corr must be from 0 to 1, but is {max_abs_corr:.2f}"

    # df_trn_val = df.loc[(df["segm"] == "train") | (df["segm"] == "val")]
    # corr_matr = df_trn_val[feature_cols].corr()
    corr_matr = df[feature_cols].corr()

    too_correlated: list[tuple[str, str]] = []
    features = feature_cols.copy()
    something_deleted = True

    while something_deleted:
        features_to_remove = []
        something_deleted = False

        for f in features:
            if f in features_to_remove or df[f].dtype == "object":
                continue

            i = corr_matr.columns.get_loc(f)
            abs_corrs = corr_matr.iloc[i, i + 1 :].abs()
            bad_corrs_cols = list(abs_corrs[abs_corrs > max_abs_corr].index)
            bad_corrs_cols = [
                g
                for g in bad_corrs_cols
                if g not in features_to_remove and g in features
            ]
            too_correlated.extend([(f, g) for g in bad_corrs_cols])
            features_to_remove.extend(bad_corrs_cols)

        if features_to_remove:
            something_deleted = True
            features = [f for f in features if f not in features_to_remove]

    return too_correlated, features


def check_train_val_test_split(
    df: pd.DataFrame,
    segm_col: str,
    target_col: str,
    treatment_col: str,
    treatment_groups_mapper: tp.Dict[tp.Any, int],
):
    # segm_col = self._base_cols_mapper["segm"]
    # target_col = self._base_cols_mapper["target"]
    # treatment_col = self._base_cols_mapper["treatment"]
    # treatment_flag = df[treatment_col].map(self._treatment_groups_mapper)

    for col in [segm_col, target_col, treatment_col]:
        assert col in df.columns

    segm = df[segm_col]
    target = df[target_col]
    treatment = df[treatment_col].map(treatment_groups_mapper)

    segm_counts = segm.value_counts()
    segm_ratios = segm.value_counts() / df.shape[0]

    # TODO: нужно ли это? если да, убрать ли одно из двух / ослабить ограничения?
    # check segment sizes
    # if segm_ratios.loc["val"] < 0.15 and segm_counts.loc["val"] < 15_000:
    #     message = f"'{segm_col}' column must contain at least 15% or 15000 of 'val' samples"
    #     raise AssertionError(message)
    # if segm_ratios.loc["test"] < 0.15 and segm_counts.loc["test"] < 15_000:
    #     message = f"'{segm_col}' column must contain at least 15% or 15000 of 'test' samples"
    #     raise AssertionError(message)

    mask_train = (segm == "train").values
    mask_val = (segm == "val").values
    mask_test = (segm == "test").values

    # check treatment = 1 ratio in segments to be equal
    treatment_train = treatment[mask_train]
    treatment_val = treatment[mask_val]
    treatment_test = treatment[mask_test]

    result_train_val = check_bernoulli_equal_means(
        treatment_train, treatment_val, alpha=0.05
    )
    result_val_test = check_bernoulli_equal_means(
        treatment_val, treatment_test, alpha=0.05
    )

    if not result_train_val["equals"] or not result_val_test["equals"]:
        raise AssertionError(
            "Treatment ratio should be the same in train, val, test splits"
        )

    # check target = 1 ratio in segments to be equal w.r.t. treatment group
    target_train_treatment = target[mask_train & (treatment == 1)]
    target_val_treatment = target[mask_val & (treatment == 1)]
    target_test_treatment = target[mask_test & (treatment == 1)]

    result_train_val = check_bernoulli_equal_means(
        target_train_treatment, target_val_treatment, alpha=0.05
    )
    result_val_test = check_bernoulli_equal_means(
        target_val_treatment, target_test_treatment, alpha=0.05
    )

    if not result_train_val["equals"] or not result_val_test["equals"]:
        raise AssertionError(
            "Target rate ratio should be the same in train, val, test splits in treatment group"
        )

    # check target = 1 ratio in segments to be equal w.r.t. control group
    target_train_control = target[mask_train & (treatment == 0)]
    target_val_control = target[mask_val & (treatment == 0)]
    target_test_control = target[mask_test & (treatment == 0)]

    result_train_val = check_bernoulli_equal_means(
        target_train_control, target_val_control, alpha=0.05
    )
    result_val_test = check_bernoulli_equal_means(
        target_val_control, target_test_control, alpha=0.05
    )

    if not result_train_val["equals"] or not result_val_test["equals"]:
        raise AssertionError(
            "Target rate ratio should be the same in train, val, test splits in control group"
        )



from .filter import FilterRanker
from .importance import ImportanceRanker
from .permutation import PermutationRanker
from .stepwise import StepwiseRanker

__all__ = [
    "FilterRanker",
    "ImportanceRanker",
    "PermutationRanker",
    "StepwiseRanker"
]



import pickle
import typing as tp
from copy import deepcopy

import pandas as pd

from ..training.gridsearch import ModelResult


class UpliftPipelineInference:
    def __init__(
        self, result: ModelResult, fillna_value: tp.Union[float, tp.Dict[str, float]]
    ):
        self.result = deepcopy(result)
        self.fillna_value = fillna_value

    def save(self, filename: str):
        with open(filename, "wb") as f:
            pickle.dump(self, f)

    @staticmethod
    def load(filename: str):
        with open(filename, "rb") as f:
            return pickle.load(f)

    def _preprocess(self, data: pd.DataFrame):
        if self.fillna_value is None:
            return
        if isinstance(self.fillna_value, dict):
            for col, value in self.fillna_value:
                data[col] = data[col].fillna(value=value)
        else:
            for col in self.result.features:
                # TODO : process categorical features
                if data[col].dtype != "object":
                    data[col] = data[col].fillna(value=self.fillna_value)

    def _postprocess(self, data: pd.DataFrame):
        if self.fillna_value is None:
            return
        if isinstance(self.fillna_value, dict):
            for col, value in self.fillna_value:
                data.loc[data[col] == value, col] = None
        else:
            for col in self.result.features:
                # TODO : process categorical features
                if data[col].dtype != "object":
                    data.loc[data[col] == self.fillna_value, col] = None

    def infer(self, data: pd.DataFrame):
        self._preprocess(data)
        uplift = self.result.estimator.predict(data[self.result.features])
        if self.result.uplift_prediction_type != "abs":
            trmnt_preds = self.result.estimator.trmnt_preds_
            ctrl_preds = self.result.estimator.ctrl_preds_
            uplift = trmnt_preds / ctrl_preds - 1
        self._postprocess(data)
        return uplift



import warnings

import numpy as np
import pandas as pd
from causalml.inference.tree import UpliftTreeClassifier

warnings.filterwarnings("ignore")


class AufTreeClassifier(UpliftTreeClassifier):
    def __init__(
        self,
        control_name,
        max_features=None,
        max_depth=3,
        min_samples_leaf=100,
        min_samples_treatment=10,
        n_reg=100,
        early_stopping_eval_diff_scale=1,
        evaluationFunction="KL",
        normalization=True,
        honesty=False,
        estimation_sample_size=0.5,
        random_state=None,
    ):

        # Сохраняем все параметры в self.params
        self.params = {
            "control_name": control_name,
            "max_features": max_features,
            "max_depth": max_depth,
            "min_samples_leaf": min_samples_leaf,
            "min_samples_treatment": min_samples_treatment,
            "n_reg": n_reg,
            "early_stopping_eval_diff_scale": early_stopping_eval_diff_scale,
            "evaluationFunction": evaluationFunction,
            "normalization": normalization,
            "honesty": honesty,
            "estimation_sample_size": estimation_sample_size,
            "random_state": random_state,
        }

        # Вызываем конструктор родительского класса
        super().__init__(**self.params)
        self.params["n_estimators"] = 1
        self.ctrl_preds_ = None
        self.trmnt_preds_ = None

    def get_params(self):
        # Возвращаем параметры
        return self.params

    def fit(self, X: pd.DataFrame, y: pd.Series, treatment: pd.Series):
        self.feature_names_ = X.columns.tolist()

        # Вызываем метод fit родительского класса
        with np.errstate(divide="ignore", invalid="ignore"):
            super().fit(
                X=X.values, y=y.values, treatment=treatment.astype("str").values
            )
        return super()

    def predict(self, X: pd.DataFrame):
        # Предсказываем uplift
        with np.errstate(divide="ignore", invalid="ignore"):
            uplift = super().predict(X.values)
        # print(uplift.shape, uplift)
        self.ctrl_preds_ = uplift[:, 0]
        # Вычисляем предсказания для обработанной группы
        self.trmnt_preds_ = uplift[:, 1]

        return uplift[:, 1] - uplift[:, 0]



import typing as tp

import numpy as np
import pandas as pd
from catboost import CatBoostClassifier
from causalml.feature_selection import FilterSelect
from causalml.inference.tree import UpliftRandomForestClassifier
from sklift.models import SoloModel


class FilterRanker:
    """Sorts features using filter-based methods for measuring the feature ability to predict uplift

    Filter means that each feature is processed separately using some procedure
    This procedure depends only on number of feature bins and divergence measure for response rates used for all bins
    This ranker sorts features at one time being able to use only train data

    Args:
        name (string): ranker name
        method (string, ['F', 'LR', 'KL', 'ED', 'Chi']): the feature selection method to be used to rank the features
            'F' for F-test
            'LR' for likelihood ratio test
            'KL', 'ED', 'Chi' for bin-based uplift filter methods, KL divergence, Euclidian distance, Chi-Square repsectively
        bins (int): number of bins to be used for bin-based uplift filter methods
        data (pd.DataFrame): DataFrame containing target, features, and treatment group
        features (1d array-like): list of feature names, that are columns in the data DataFrame
        target_col (string): target column name
        treatment_col (string): treatment column name

    Examples:
        >>> from AUF.feature_selection import FilterRanker
        >>> ranker = FilterRanker(method = "KL", bins = 10)
        >>> ranked_features, ranked_importances = ranker.run(df, features, target_col, treatment_col)

    Returns: None
    """

    def __init__(self, method: str, bins: int, name: str = "__empty_name__"):
        self.name: str = name
        self.method: str = method
        self.bins: int = bins
        self.ranked_features: tp.List[str] = []
        self.ranked_features_scores: tp.List[float] = []

    def run(
        self,
        data: pd.DataFrame,
        features: tp.List[str],
        target_col: str,
        treatment_col: str,
    ) -> tp.List[str]:

        assert "treatment_group_key" not in data.columns
        assert ((data[treatment_col] == 0) | (data[treatment_col] == 1)).all()

        try:
            data["treatment_group_key"] = data[treatment_col].apply(
                lambda x: "control" if x == 0 else "treatment"
            )

            selector = FilterSelect()

            result = selector.filter_D(
                data=data,
                features=features,
                y_name=target_col,
                n_bins=self.bins,
                method=self.method,
                null_impute="mean",  # {'mean', 'median', 'most_frequent', None}
                experiment_group_column="treatment_group_key",
                control_group="control",
            )

        except Exception as e:
            print("Exception was raised:", str(e))
            raise e

        finally:
            data.drop("treatment_group_key", axis=1, inplace=True)

        self.ranked_features = list(result.feature.values)
        self.ranked_features_scores = list(result.score.values)
        return self.ranked_features, self.ranked_features_scores

    def get_ranked_features(self):
        return self.ranked_features

    def get_ranked_features_scores(self):
        return self.ranked_features_scores

    def get_ranker_name(self) -> str:
        return self.name



import sys
import subprocess
from setuptools import setup, find_packages

if __name__ == '__main__':
    # Предустановка lightgbm перед основной установкой
    subprocess.check_call([
        sys.executable, "-m", "pip", "install",
        "lightgbm>=3.3.5",
        "--config-settings=cmake.define.USE_OPENMP=OFF"
    ])
    setup()



from .models import (
    RANDOM_STATE,
    BOOTSTRAP_REPEATS,
    BASE_MODEL,
    SOLO_MODEL,
    TWO_MODEL,
    UPLIFT_TREE,
    UPLIFT_FOREST,
    X_LEARNER
)

from .metrics import (
    METRICS,
    METRICS_UPLIFT_AT_K,
    METRICS_UPLIFT_REL_AT_K,
    METRICS_WEIGHTED_UPLIFT,
    METRICS_WEIGHTED_UPLIFT_STABLE,
    METRICS_WEIGHTED_UPLIFT_REL,
    METRICS_WEIGHTED_UPLIFT_REL_STABLE,
    METRICS_CONTROL_TREATMENT_TARGET_RATIOS_AT_K,
    METRICS_ABS_UPLIFT_GROWTH_AT_K,
    METRICS_REL_UPLIFT_GROWTH_AT_K,
    METRICS_ABS_REL_UPLIFT_GROWTH_AT_K,
    METRICS_OTHERS,
    METRICS_OTHERS_MY
)

__all__ = [
    "METRICS",
    "METRICS_UPLIFT_AT_K",
    "METRICS_UPLIFT_REL_AT_K",
    "METRICS_WEIGHTED_UPLIFT",
    "METRICS_WEIGHTED_UPLIFT_STABLE",
    "METRICS_WEIGHTED_UPLIFT_REL",
    "METRICS_WEIGHTED_UPLIFT_REL_STABLE",
    "METRICS_CONTROL_TREATMENT_TARGET_RATIOS_AT_K",
    "METRICS_ABS_UPLIFT_GROWTH_AT_K",
    "METRICS_REL_UPLIFT_GROWTH_AT_K",
    "METRICS_ABS_REL_UPLIFT_GROWTH_AT_K",
    "METRICS_OTHERS",
    "METRICS_OTHERS_M",
    "RANDOM_STATE",
    "BOOTSTRAP_REPEATS",
    "BASE_MODEL",
    "SOLO_MODEL",
    "TWO_MODEL",
    "UPLIFT_TREE",
    "UPLIFT_FOREST",
    "X_LEARNER"
]



import typing as tp
from dataclasses import dataclass

import numpy as np
import optuna
import pandas as pd
from catboost import CatBoostClassifier, CatBoostRegressor
from causalml.inference.tree import UpliftRandomForestClassifier
from sklearn.pipeline import Pipeline
from sklift.models import SoloModel, TwoModels

from ..constants import (
    RANDOM_STATE,
    SOLO_MODEL,
    TWO_MODEL,
    UPLIFT_FOREST,
    UPLIFT_TREE,
    X_LEARNER,
)
from ..models import AufRandomForestClassifier, AufTreeClassifier, AufXLearner

# from functools import partical


@dataclass
class ModelResult:
    estimator: Pipeline
    features: tp.List[str]
    uplift_prediction_type: str  # same on training & prediction
    median_test_metric_value: float


class OptuneUpliftDefault:
    "Find the best parameters for a given model class using a user-supplied metric for optimization"

    def __init__(
        self,
        df_train: pd.DataFrame,
        df_valid: pd.DataFrame,
        metric: object,
        treatment_col: str,
        target_col: str,
        overfit_metric: object,
        uplift_type: str,
    ):
        self.df_train = df_train
        self.df_valid = df_valid
        self.metric = metric
        self.treatment_col = treatment_col
        self.target_col = target_col
        self.overfit_metric = overfit_metric
        self.uplift_type = uplift_type

    def objective_slearner(self, trial):
        param = {
            "objective": trial.suggest_categorical(
                "objective", ["Logloss", "CrossEntropy"]
            ),
            "depth": trial.suggest_int("depth", 1, 8),
            "learning_rate": trial.suggest_float("learning_rate", 0.01, 0.3),
            "n_estimators": trial.suggest_int("n_estimators", 30, 240),
            "boosting_type": trial.suggest_categorical(
                "boosting_type", ["Ordered", "Plain"]
            ),
            "bootstrap_type": trial.suggest_categorical(
                "bootstrap_type", ["Bayesian", "Bernoulli", "MVS"]
            ),
        }

        if param["bootstrap_type"] == "Bayesian":
            param["bagging_temperature"] = trial.suggest_float(
                "bagging_temperature", 0, 10
            )
        elif param["bootstrap_type"] == "Bernoulli":
            param["subsample"] = trial.suggest_float("subsample", 0.1, 1)
        solo_method = trial.suggest_categorical(
            "method", ["dummy", "treatment_interaction"]
        )

        # Создание и обучение модели
        estimator = CatBoostClassifier(random_state=RANDOM_STATE, silent=True, **param)
        model = self.model_class(method=solo_method, estimator=estimator)

        model.fit(
            X=self.df_train[self.features],
            y=self.df_train[self.target_col],
            treatment=self.df_train[self.treatment_col],
        )
        uplift_valid = model.predict(self.df_valid[self.features])
        if self.uplift_type == "rel":
            uplift_valid = model.trmnt_preds_ / model.ctrl_preds_

        metric_result = self.metric(
            y_true=self.df_valid[self.target_col],
            uplift=uplift_valid,
            treatment=self.df_valid[self.treatment_col],
        )
        if self.overfit_metric is not None:
            uplift_train = model.predict(self.df_train[self.features])
            if self.uplift_type == "rel":
                uplift_train = model.trmnt_preds_ / model.ctrl_preds_

            metric_result_train = self.metric(
                y_true=self.df_train[self.target_col],
                uplift=uplift_train,
                treatment=self.df_train[self.treatment_col],
            )
            return self.overfit_metric(metric_result, metric_result_train)

        return metric_result

    def objective_tlearner(self, trial):
        param = {
            "objective": trial.suggest_categorical(
                "objective", ["Logloss", "CrossEntropy"]
            ),
            "depth": trial.suggest_int("depth", 1, 8),
            "learning_rate": trial.suggest_float("learning_rate", 0.01, 0.3),
            "n_estimators": trial.suggest_int("n_estimators", 30, 240),
            "boosting_type": trial.suggest_categorical(
                "boosting_type", ["Ordered", "Plain"]
            ),
            "bootstrap_type": trial.suggest_categorical(
                "bootstrap_type", ["Bayesian", "Bernoulli", "MVS"]
            ),
        }

        if param["bootstrap_type"] == "Bayesian":
            param["bagging_temperature"] = trial.suggest_float(
                "bagging_temperature", 0, 10
            )
        elif param["bootstrap_type"] == "Bernoulli":
            param["subsample"] = trial.suggest_float("subsample", 0.1, 1)

        param2 = {
            "objective": trial.suggest_categorical(
                "objective2", ["Logloss", "CrossEntropy"]
            ),
            "depth": trial.suggest_int("depth2", 1, 8),
            "learning_rate": trial.suggest_float("learning_rate2", 0.01, 0.3),
            "n_estimators": trial.suggest_int("n_estimators2", 30, 240),
            "boosting_type": trial.suggest_categorical(
                "boosting_type2", ["Ordered", "Plain"]
            ),
            "bootstrap_type": trial.suggest_categorical(
                "bootstrap_type2", ["Bayesian", "Bernoulli", "MVS"]
            ),
        }

        if param2["bootstrap_type"] == "Bayesian":
            param2["bagging_temperature"] = trial.suggest_float(
                "bagging_temperature2", 0, 10
            )
        elif param2["bootstrap_type"] == "Bernoulli":
            param2["subsample"] = trial.suggest_float("subsample2", 0.1, 1)
        two_method = trial.suggest_categorical(
            "method", ["vanilla", "ddr_control", "ddr_treatment"]
        )

        # Создание и обучение модели
        estimator = CatBoostClassifier(random_state=RANDOM_STATE, silent=True, **param)
        estimator2 = CatBoostClassifier(
            random_state=RANDOM_STATE, silent=True, **param2
        )
        model = self.model_class(
            method=two_method, estimator_trmnt=estimator, estimator_ctrl=estimator2
        )

        model.fit(
            X=self.df_train[self.features],
            y=self.df_train[self.target_col],
            treatment=self.df_train[self.treatment_col],
        )

        uplift_valid = model.predict(self.df_valid[self.features])
        if self.uplift_type == "rel":
            uplift_valid = model.trmnt_preds_ / model.ctrl_preds_

        metric_result = self.metric(
            y_true=self.df_valid[self.target_col],
            uplift=uplift_valid,
            treatment=self.df_valid[self.treatment_col],
        )
        if self.overfit_metric is not None:
            uplift_train = model.predict(self.df_train[self.features])
            if self.uplift_type == "rel":
                uplift_train = model.trmnt_preds_ / model.ctrl_preds_

            metric_result_train = self.metric(
                y_true=self.df_train[self.target_col],
                uplift=uplift_train,
                treatment=self.df_train[self.treatment_col],
            )
            return self.overfit_metric(metric_result, metric_result_train)

        return metric_result

    def objective_tree(self, trial):
        param = {
            "max_depth": trial.suggest_int("max_depth", 1, 10),
            "max_features": trial.suggest_int(
                "max_features", 2, int(max(np.sqrt(len(self.features)), 3))
            ),
            "estimation_sample_size": trial.suggest_float(
                "estimation_sample_size", 0.1, 0.9
            ),
            "evaluationFunction": trial.suggest_categorical(
                "evaluationFunction",
                ["KL", "ED", "Chi", "CTS", "DDP", "IT", "CIT", "IDDP"],
            ),
        }
        model = self.model_class(control_name="0", random_state=RANDOM_STATE, **param)
        model.fit(
            X=self.df_train[self.features],
            y=self.df_train[self.target_col],
            treatment=self.df_train[self.treatment_col],
        )
        uplift_valid = model.predict(self.df_valid[self.features])

        metric_result = self.metric(
            y_true=self.df_valid[self.target_col],
            uplift=uplift_valid,
            treatment=self.df_valid[self.treatment_col],
        )

        if self.overfit_metric is not None:
            uplift_train = model.predict(self.df_train[self.features])

            metric_result_train = self.metric(
                y_true=self.df_train[self.target_col],
                uplift=uplift_train,
                treatment=self.df_train[self.treatment_col],
            )
            return self.overfit_metric(metric_result, metric_result_train)

        return metric_result

    def objective_random_forest(self, trial):
        param = {
            "max_depth": trial.suggest_int("max_depth", 1, 4),
            "max_features": trial.suggest_int(
                "max_features", 2, int(max(np.sqrt(len(self.features)), 3))
            ),
            "n_estimators": trial.suggest_int("n_estimators", 30, 100),
            "estimation_sample_size": trial.suggest_float(
                "estimation_sample_size", 0.2, 0.8
            ),
            "evaluationFunction": trial.suggest_categorical(
                "evaluationFunction",
                ["KL", "ED", "Chi", "CTS", "DDP", "IT", "CIT", "IDDP"],
            ),
        }
        model = self.model_class(control_name="0", random_state=RANDOM_STATE, **param)
        model.fit(
            X=self.df_train[self.features],
            y=self.df_train[self.target_col],
            treatment=self.df_train[self.treatment_col],
        )
        uplift_valid = model.predict(self.df_valid[self.features])

        metric_result = self.metric(
            y_true=self.df_valid[self.target_col],
            uplift=uplift_valid,
            treatment=self.df_valid[self.treatment_col],
        )

        if self.overfit_metric is not None:
            uplift_train = model.predict(self.df_train[self.features])

            metric_result_train = self.metric(
                y_true=self.df_train[self.target_col],
                uplift=uplift_train,
                treatment=self.df_train[self.treatment_col],
            )
            return self.overfit_metric(metric_result, metric_result_train)

        return metric_result

    def objective_catboost(self, trial):
        param = {
            "objective": trial.suggest_categorical(
                "objective", ["Logloss", "CrossEntropy"]
            ),
            "depth": trial.suggest_int("depth", 1, 8),
            "learning_rate": trial.suggest_float("learning_rate", 0.01, 0.3),
            "n_estimators": trial.suggest_int("n_estimators", 30, 240),
            "boosting_type": trial.suggest_categorical(
                "boosting_type", ["Ordered", "Plain"]
            ),
            "bootstrap_type": trial.suggest_categorical(
                "bootstrap_type", ["Bayesian", "Bernoulli", "MVS"]
            ),
        }

        if param["bootstrap_type"] == "Bayesian":
            param["bagging_temperature"] = trial.suggest_float(
                "bagging_temperature", 0, 10
            )
        elif param["bootstrap_type"] == "Bernoulli":
            param["subsample"] = trial.suggest_float("subsample", 0.1, 1)

        # Создание и обучение модели
        estimator = CatBoostClassifier(random_state=RANDOM_STATE, silent=True, **param)

        estimator.fit(
            X=self.df_train[self.features],
            y=self.df_train[self.target_col],
        )

        score_vld = estimator.predict_proba(self.df_valid[self.features])[:, 1]
        try:
            metric_result = self.metric(
                y_true=self.df_valid[self.target_col],
                y_score=score_vld,
            )
            if self.overfit_metric is not None:
                score_train = estimator.predict_proba(self.df_train[self.features])[
                    :, 1
                ]

                metric_result_train = self.metric(
                    y_true=self.df_train[self.target_col],
                    y_score=score_train,
                )
                return self.overfit_metric(metric_result, metric_result_train)
        except:
            metric_result = self.metric(
                y_true=self.df_valid[self.target_col],
                y_pred=(score_vld >= 0.5).astype("int"),
            )
            if self.overfit_metric is not None:
                score_train = estimator.predict_proba(self.df_train[self.features])[
                    :, 1
                ]

                metric_result_train = self.metric(
                    y_true=self.df_train[self.target_col],
                    y_pred=(score_train >= 0.5).astype("int"),
                )
                return self.overfit_metric(metric_result, metric_result_train)
        return metric_result

    def objective_x_learner(self, trial):
        # param grid for the first step model of x-learner
        # CatBoostClassifier
        model_params = {
            "objective": trial.suggest_categorical(
                "objective", ["Logloss", "CrossEntropy"]
            ),
            "depth": trial.suggest_int("depth", 1, 8),
            "learning_rate": trial.suggest_float("learning_rate", 0.01, 0.3),
            "n_estimators": trial.suggest_int("n_estimators", 30, 240),
            "boosting_type": trial.suggest_categorical(
                "boosting_type", ["Ordered", "Plain"]
            ),
            "bootstrap_type": trial.suggest_categorical(
                "bootstrap_type", ["Bayesian", "Bernoulli", "MVS"]
            ),
        }

        if model_params["bootstrap_type"] == "Bayesian":
            model_params["bagging_temperature"] = trial.suggest_float(
                "bagging_temperature", 0, 10
            )
        elif model_params["bootstrap_type"] == "Bernoulli":
            model_params["subsample"] = trial.suggest_float("subsample", 0.1, 1)

        # param grid for the second step model of x-learner
        # CatBoostRegressor
        uplift_model_params = {
            "objective": trial.suggest_categorical(
                "objective2", ["MAE", "MAPE", "RMSE"]
            ),
            "depth": trial.suggest_int("depth2", 1, 8),
            "learning_rate": trial.suggest_float("learning_rate2", 0.01, 0.3),
            "n_estimators": trial.suggest_int("n_estimators2", 30, 240),
            "boosting_type": trial.suggest_categorical(
                "boosting_type2", ["Ordered", "Plain"]
            ),
            "bootstrap_type": trial.suggest_categorical(
                "bootstrap_type2", ["Bayesian", "Bernoulli", "MVS"]
            ),
        }

        if uplift_model_params["bootstrap_type"] == "Bayesian":
            uplift_model_params["bagging_temperature"] = trial.suggest_float(
                "bagging_temperature2", 0, 10
            )
        elif uplift_model_params["bootstrap_type"] == "Bernoulli":
            uplift_model_params["subsample"] = trial.suggest_float("subsample2", 0.1, 1)

        # create base models and x-learner
        model = CatBoostClassifier(
            random_state=RANDOM_STATE, silent=True, **model_params
        )
        uplift_model = CatBoostRegressor(
            random_state=RANDOM_STATE, silent=True, **uplift_model_params
        )
        x_learner = self.model_class(
            model=model,
            uplift_model=uplift_model,
            map_groups={"control": 0, "treatment": 1},
        )

        x_learner.fit(
            X=self.df_train[self.features],
            y=self.df_train[self.target_col],
            treatment=self.df_train[self.treatment_col],
        )

        uplift_valid = x_learner.predict(self.df_valid[self.features])

        # TODO : think about relative uplift for x_learner
        #        1) 2nd step = uplift regressors --> division not needed
        #        2) 1st step = propensities --> division = usual two-models
        # if self.uplift_type == 'rel':
        #     uplift_valid = x_learner.trmnt_preds_ / model.ctrl_preds_

        metric_result = self.metric(
            y_true=self.df_valid[self.target_col],
            uplift=uplift_valid,
            treatment=self.df_valid[self.treatment_col],
        )
        if self.overfit_metric is not None:
            uplift_train = x_learner.predict(self.df_train[self.features])

            # TODO : look at previous todo
            # if self.uplift_type == 'rel':
            #     uplift_train = model.trmnt_preds_ / model.ctrl_preds_

            metric_result_train = self.metric(
                y_true=self.df_train[self.target_col],
                uplift=uplift_train,
                treatment=self.df_train[self.treatment_col],
            )
            return self.overfit_metric(metric_result, metric_result_train)

        return metric_result

    def objective_custom(self, trial):
        """
        Custom  object function for custom class

        Args:
            optuna trial

        Returns: None
        """
        assert (
            -1 > 1
        ), "Для использования своего objective для своего класса модели определите objective_custom"

    def find_best_params(
        self, model_class: Pipeline, features: tp.List[str], timeout: int
    ):
        """
        Finds the optimal parameters for the given model class to maximize the specified function

        Args:
            model_class (pipeline): the model class for which the optimal parameters
                will be found. It should be a scikit-learn compatible pipeline or model

            features (1d array-like): a list of feature names to be used in the model. These features
                will be considered during the parameter optimization process

            timeout (int): the maximum time in seconds allowed for the optimization
                process. If the optimization exceeds this time, it will be terminated

        Returns: estimator with optimal values
        """
        optuna.logging.set_verbosity(optuna.logging.WARNING)
        sampler = optuna.samplers.TPESampler(seed=RANDOM_STATE)
        study = optuna.create_study(direction="maximize", sampler=sampler)
        self.model_class = model_class
        self.features = features

        if model_class.__name__ == "SoloModel":
            study.optimize(self.objective_slearner, timeout=timeout)
            params = study.best_params
            method = params["method"]
            del params["method"]
            estimator = CatBoostClassifier(
                random_state=RANDOM_STATE, silent=True, **params
            )
            slearner = model_class(method=method, estimator=estimator)
            return slearner

        elif model_class.__name__ == "TwoModels":
            study.optimize(self.objective_tlearner, timeout=timeout)
            params = study.best_params
            method = params["method"]
            del params["method"]
            param1 = {k: v for k, v in params.items() if k[-1] != "2"}
            param2 = {k[:-1]: v for k, v in params.items() if k[-1] == "2"}
            estimator = CatBoostClassifier(
                random_state=RANDOM_STATE, silent=True, **param1
            )
            estimator2 = CatBoostClassifier(
                random_state=RANDOM_STATE, silent=True, **param2
            )
            tlearner = model_class(
                method=method, estimator_trmnt=estimator, estimator_ctrl=estimator2
            )
            return tlearner

        elif model_class.__name__ == "AufRandomForestClassifier":
            study.optimize(self.objective_random_forest, timeout=timeout)
            params = study.best_params
            forest = model_class(control_name="0", random_state=RANDOM_STATE, **params)
            return forest

        elif model_class.__name__ == "AufTreeClassifier":
            study.optimize(self.objective_tree, timeout=timeout)
            params = study.best_params
            tree = model_class(control_name="0", random_state=RANDOM_STATE, **params)
            return tree

        elif model_class.__name__ == "CatBoostClassifier":
            study.optimize(self.objective_catboost, timeout=timeout)
            params = study.best_params
            response_model = model_class(
                random_state=RANDOM_STATE, silent=True, **params
            )
            return response_model

        elif model_class.__name__ == "AufXLearner":
            study.optimize(self.objective_x_learner, timeout=timeout)
            params = study.best_params
            model_params = {k: v for k, v in params.items() if k[-1] != "2"}
            uplift_model_params = {k[:-1]: v for k, v in params.items() if k[-1] == "2"}
            model = CatBoostClassifier(
                random_state=RANDOM_STATE, silent=True, **model_params
            )
            uplift_model = CatBoostRegressor(
                random_state=RANDOM_STATE, silent=True, **uplift_model_params
            )
            x_learner = model_class(
                model=model,
                uplift_model=uplift_model,
                map_groups={"control": 0, "treatment": 1},
            )
            return x_learner

        else:
            study.optimize(self.objective_custom, timeout=timeout)
            params = study.best_params
            forest = model_class(**params)
            return forest



from . import constants
from . import data
from . import feature_rankers
from . import metrics
from . import ml_flow
from . import models
from . import pipeline
from . import plots
from . import training



__version__ = "1.0.0"
__all__ = [
    "constants",
    "data",
    "feature_rankers",
    "metrics",
    "ml_flow",
    "models",
    "pipeline",
    "plots",
    "training"
]